"""
IO methods to parse a variety of file formats.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import warnings
from string import ascii_lowercase
from dataclasses import dataclass
import xml.etree.ElementTree as ET
from typing import List, Dict, Any, Optional

import numpy as np
from scipy.spatial.transform import Rotation

from .. import meshing
from ..utils import volume_to_points, compute_bounding_box, NORMAL_REFERENCE


def _parse_data_array(data_array: ET.Element, dtype: type = float) -> np.ndarray:
    """
    Parse a DataArray element into a numpy array.

    Parameters
    ----------
    data_array : ET.Element
        XML element containing array data.
    dtype : type, optional
        Data type for parsing, by default float.

    Returns
    -------
    np.ndarray
        Parsed numpy array.
    """
    rows = [row.strip() for row in data_array.text.strip().split("\n") if row.strip()]
    parsed_rows = [[dtype(x) for x in row.split()] for row in rows]
    data = np.array(parsed_rows)
    return np.squeeze(data)


def _parse_dtype(xml_element) -> object:
    """
    Determine data type from XML element type attribute.

    Parameters
    ----------
    xml_element : ET.Element
        XML element to parse type from.

    Returns
    -------
    object
        Data type (float or int).
    """
    return float if xml_element.get("type", "").startswith("Float") else int


@dataclass
class GeometryData:
    """
    Container for single geometry entity data.

    Parameters
    ----------
    vertices : np.ndarray, optional
        3D vertex coordinates.
    normals : np.ndarray, optional
        Normal vectors at each vertex.
    faces : np.ndarray, optional
        Face connectivity indices.
    quaternions : np.ndarray, optional
        Orientation quaternions for each vertex.
    vertex_properties : VertexPropertyContainer, optional
        Additional vertex properties.
    shape : List[int], optional
        Bounding box dimensions.
    sampling : List[float], optional
        Sampling rates along each axis, by default (1, 1, 1).
    """

    vertices: np.ndarray = None
    normals: np.ndarray = None
    faces: np.ndarray = None
    quaternions: np.ndarray = None
    vertex_properties: "VertexPropertyContainer" = None
    shape: List[int] = None
    sampling: List[float] = (1, 1, 1)


@dataclass
class GeometryDataContainer:
    """
    Container for multiple geometry entities with automatic validation.

    Parameters
    ----------
    vertices : List[np.ndarray], optional
        List of vertex arrays for each geometry entity.
    normals : List[np.ndarray], optional
        List of normal arrays for each geometry entity.
    faces : List[np.ndarray], optional
        List of face arrays for each geometry entity.
    quaternions : List[np.ndarray], optional
        List of quaternion arrays for each geometry entity.
    vertex_properties : List[VertexPropertyContainer], optional
        List of vertex property containers for each geometry entity.
    shape : List[int], optional
        Bounding box dimensions.
    sampling : List[float], optional
        Sampling rates along each axis, by default (1, 1, 1).
    """

    vertices: List[np.ndarray] = None
    normals: List[np.ndarray] = None
    faces: List[np.ndarray] = None
    quaternions: List[np.ndarray] = None
    vertex_properties: List["VertexPropertyContainer"] = None
    shape: List[int] = None
    sampling: List[float] = (1, 1, 1)

    def __post_init__(self):
        dtype_map = {
            "vertices": np.float32,
            "normals": np.float32,
            "faces": int,
            "quaternions": np.float32,
        }

        if self.normals is None:
            self.normals = [None for x in self.vertices]

        for i in range(len(self.normals)):
            if self.normals[i] is None:
                continue
            norm = np.linalg.norm(self.normals[i], axis=1)
            mask = norm < 1e-12
            norm[mask] = 1
            self.normals[i][mask] = NORMAL_REFERENCE
            self.normals[i] = self.normals[i] / norm[:, None]

        if self.quaternions is None:
            self.quaternions = [None for x in self.vertices]

        if self.vertex_properties is None:
            self.vertex_properties = [VertexPropertyContainer() for _ in self.vertices]

        for attr_name, dtype in dtype_map.items():
            attr = getattr(self, attr_name)
            setattr(self, attr_name, self._to_dtype(attr, dtype))

        if self.shape is None:
            self.shape, _ = compute_bounding_box(self.vertices)

        if len(self.vertices) != len(self.normals):
            raise ValueError("Normals need to be specified for each vertex set.")

        if self.faces is not None:
            if len(self.vertices) != len(self.faces):
                raise ValueError("Faces need to be specified for each vertex set.")

    def __len__(self):
        return len(self.vertices)

    def __iter__(self):
        yield from [self[i] for i in range(len(self))]

    def __getitem__(self, index: int) -> GeometryData:
        return GeometryData(
            vertices=self.vertices[index],
            normals=self.normals[index],
            shape=self.shape,
            sampling=self.sampling,
            faces=self.faces[index] if self.faces is not None else None,
            quaternions=self.quaternions[index],
            vertex_properties=self.vertex_properties[index],
        )

    @staticmethod
    def _to_dtype(data: List[np.ndarray], dtype=np.float32):
        try:
            n_elements = len(data)
        except Exception:
            n_elements = 0

        for i in range(n_elements):
            try:
                data[i] = data[i].astype(dtype)
            except Exception:
                pass
        return data


class VertexPropertyContainer:
    """
    Container for managing custom vertex properties with automatic synchronization.

    Parameters
    ----------
    properties : dict of str -> np.ndarray, optional
        Dictionary mapping property names to vertex data arrays.
    """

    def __init__(self, properties: Optional[Dict[str, np.ndarray]] = None):
        """
        Initialize vertex property container.

        Parameters
        ----------
        properties : dict of str -> np.ndarray, optional
            Dictionary mapping property names to vertex data arrays
        """
        properties = {} if properties is None else properties
        properties = {name: np.asarray(data) for name, data in properties.items()}

        # We use len instead of size for future vector field support
        self._n_vertices = max((*(len(x) for x in properties.values()), 0))
        for name, data in properties.items():
            if len(data) == self._n_vertices:
                continue
            raise ValueError(
                f"Property '{name}' has {len(data)} values, "
                f"but expected {self._n_vertices} to match vertex count"
            )
        self._properties = properties

    def __getitem__(self, idx: str) -> "VertexPropertyContainer":
        """Array-like indexing using int/bool numpy arrays, slices or ellipses."""
        if not self._properties:
            return VertexPropertyContainer()

        if isinstance(idx, (int, np.integer)):
            idx = [idx]
        elif isinstance(idx, slice) or idx is ...:
            idx = np.arange(self._n_vertices)[idx]

        idx = np.asarray(idx)
        if idx.dtype == bool:
            idx = np.where(idx)[0]

        return VertexPropertyContainer(
            {k: v[idx].copy() for k, v in self._properties.items()}
        )

    @property
    def properties(self):
        """List available vertex properties."""
        return list(self._properties.keys())

    def get_property(self, name: str, default: Any = None) -> Optional[np.ndarray]:
        """Get property data by name."""
        return self._properties.get(name, default)

    def remove_property(self, name: str) -> None:
        _ = self._properties.pop(name, None)

    def set_property(self, name: str, data: np.ndarray) -> None:
        """Insert or replace a vertex property.

        Parameters
        ----------
        name : str
            Property name.
        data : np.ndarray
            Array of per-vertex values.

        Raises
        ------
        ValueError
            If *data* length does not match the existing vertex count.
        """
        data = np.asarray(data)
        if self._properties and len(data) != self._n_vertices:
            raise ValueError(
                f"Property '{name}' has {len(data)} values, "
                f"but expected {self._n_vertices} to match vertex count"
            )
        if not self._properties:
            self._n_vertices = len(data)
        self._properties[name] = data

    def copy(self) -> "VertexPropertyContainer":
        """Create a deep copy of the container."""
        return self[...]

    @classmethod
    def merge(
        cls, containers: List["VertexPropertyContainer"]
    ) -> "VertexPropertyContainer":
        """
        Merge multiple property containers.

        Parameters
        ----------
        containers : list of VertexPropertyContainer
            Containers to merge

        Returns
        -------
        VertexPropertyContainer
            New container with merged properties
        """
        containers = [c for c in containers if c._properties]
        if not containers:
            return cls()

        all_props = set(containers[0].properties)
        common_props = set(containers[0].properties)
        for container in containers[1:]:
            container_props = set(container.properties)

            common_props &= container_props
            all_props |= container_props

        if not common_props:
            warnings.warn("No common properties found across containers to merge")
            return cls()

        dropped_props = all_props - common_props
        if dropped_props:
            warnings.warn(
                f"Properties {sorted(dropped_props)} were not common across all "
                f"containers and were dropped during merge"
            )

        merged_props = {}
        for prop_name in common_props:
            merged_props[prop_name] = np.concatenate(
                [container.get_property(prop_name) for container in containers], axis=0
            )
        return cls(merged_props)


def _read_orientations(filename: str):
    """
    Read orientation data from file and convert to geometry format.

    Parameters
    ----------
    filename : str
        Path to orientation file.

    Returns
    -------
    dict
        Dictionary containing vertices, normals, and quaternions.
    """
    from tme import Orientations

    data = Orientations.from_file(filename)

    # Remap as active (push) rotation
    angles = Rotation.from_euler(seq="ZYZ", angles=data.rotations, degrees=True).inv()

    normals = angles.apply(NORMAL_REFERENCE)
    quaternions = angles.as_quat(scalar_first=True)
    indices = [np.where(data.details == x) for x in np.unique(data.details)]

    try:
        vertex_properties = [
            VertexPropertyContainer(
                {
                    "pytme_score": data.scores[x],
                    "entity": data.details[x],
                }
            )
            for x in indices
        ]
    except Exception:
        vertex_properties = None

    return {
        "vertices": [data.translations[x] for x in indices],
        "normals": [normals[x] for x in indices],
        "quaternions": [quaternions[x] for x in indices],
        "vertex_properties": vertex_properties,
    }


def read_star(filename: str):
    """
    Read RELION star file format.

    Parameters
    ----------
    filename : str
        Path to star file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    return GeometryDataContainer(**_read_orientations(filename))


def read_txt(filename: str):
    """
    Read text-based point cloud files.

    Parameters
    ----------
    filename : str
        Path to text file (txt, csv, xyz).

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    ret = []

    delimiter = None
    if filename.endswith(("csv", "xyz")):
        delimiter = ","
    elif filename.endswith(("txt", "tsv")):
        delimiter = "\t"

    with open(filename, mode="r") as ifile:
        data = ifile.read().split("\n")
        data = [x.strip().split(delimiter) for x in data if x.strip()]

    header = ("x", "y", "z", *ascii_lowercase)[: len(data[0])]
    if "x" in data[0]:
        header = data.pop(0)

    required_columns = ("x", "y", "z")
    for rc in required_columns:
        if rc in header:
            continue
        raise ValueError(f"Colums {required_columns} are required.")

    data = {c: np.asarray(d) for c, d in zip(header, zip(*data))}

    if "id" in data:
        ret = []
        for cluster in np.unique(data["id"]):
            ret.append({c: d[data["id"] == cluster] for c, d in data.items()})
        data = ret
    else:
        data = [data]

    vertices, normals, quaternions = [], [], []
    for cluster in data:
        cols = ("x", "y", "z")
        vertices.append((np.hstack([cluster[k][:, None] for k in cols])))
        try:
            cols = ("nx", "ny", "nz")
            normals.append((np.hstack([cluster[k][:, None] for k in cols])))
        except Exception as e:
            continue

    if len(normals) == 0:
        normals = None

    return GeometryDataContainer(vertices=vertices, normals=normals)


def read_tsv(filename: str) -> GeometryDataContainer:
    """
    Read tab-separated values file with orientation data.

    Parameters
    ----------
    filename : str
        Path to tsv file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    with open(filename, mode="r") as infile:
        header = infile.readline()
    if "euler" not in header:
        return read_txt(filename)
    return GeometryDataContainer(**_read_orientations(filename))


def read_tsi(filename: str) -> GeometryDataContainer:
    """
    Read topology surface information file format.

    Parameters
    ----------
    filename : str
        Path to tsi file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    data = _read_tsi_file(filename)
    mesh = meshing.utils.to_open3d(data["vertices"][:, 1:4], data["faces"][:, 1:4])
    vertex_properties = {}

    try:
        if "inclusions" in data:
            inclusions = np.zeros((len(data["vertices"])))
            inclusion_type = data["inclusions"][:, 1]
            inclusion_vert = data["inclusions"][:, 2].astype(int)
            inclusions[inclusion_vert] = inclusion_type
            vertex_properties = {"inclusion": inclusions}
    except Exception:
        pass
    return _return_mesh(mesh, vertex_properties=vertex_properties)


def read_vtu(filename: str) -> GeometryDataContainer:
    """
    Read VTK unstructured grid XML file format.

    Parameters
    ----------
    filename : str
        Path to vtu file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    data = _read_vtu_file(filename)
    mesh = meshing.utils.to_open3d(data["points"], data["connectivity"])
    return _return_mesh(mesh, vertex_properties=data.get("point_data", {}))


def read_mesh(filename: str) -> GeometryDataContainer:
    """
    Read 3D mesh files using Open3D.

    Parameters
    ----------
    filename : str
        Path to mesh file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    import open3d as o3d

    return _return_mesh(o3d.io.read_triangle_mesh(filename))


def _return_mesh(mesh, vertex_properties: dict = None) -> GeometryDataContainer:
    """
    Convert Open3D mesh to GeometryDataContainer.

    Parameters
    ----------
    mesh : o3d.geometry.TriangleMesh
        Open3D triangle mesh object.
    vertex_properties : dict, optional
        Vertex property data.

    Returns
    -------
    GeometryDataContainer
        Converted geometry data container.
    """
    if not mesh.has_vertex_normals():
        mesh.compute_vertex_normals()

    return GeometryDataContainer(
        vertices=[np.asarray(mesh.vertices)],
        faces=[np.asarray(mesh.triangles)],
        normals=[np.asarray(mesh.vertex_normals)],
        vertex_properties=[VertexPropertyContainer(vertex_properties)],
    )


def read_structure(filename: str) -> GeometryDataContainer:
    """
    Read molecular structure files.

    Parameters
    ----------
    filename : str
        Path to structure file (pdb, cif, gro).

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    from tme import Structure

    data = Structure.from_file(filename)
    return GeometryDataContainer(vertices=[data.atom_coordinate])


def read_volume(filename: str):
    """
    Read 3D volume data and convert to point cloud.

    Parameters
    ----------
    filename : str
        Path to volume file.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.
    """
    volume = load_density(filename)

    shape = np.multiply(volume.shape, volume.sampling_rate)
    ret = volume_to_points(
        volume.data, volume.sampling_rate, reverse_order=False, max_cluster=10000
    )
    return GeometryDataContainer(
        vertices=ret, shape=shape, sampling=volume.sampling_rate
    )


def _read_tsi_file(file_path: str) -> Dict:
    """
    Reads a topology file [1]_.

    Parameters
    ----------
    file_path : str
        The path to the topology file to be parsed.

    Returns
    -------
    Dict
        Topology file content.

    References
    ----------
    .. [1] https://github.com/weria-pezeshkian/FreeDTS/wiki/Manual-for-version-1
    """
    from ._utils import _drop_prefix

    _keys = ("version", "box", "n_vertices", "vertices", "n_faces", "faces")
    ret = {k: None for k in _keys}

    with open(file_path, mode="r", encoding="utf-8") as infile:
        data = [x.strip() for x in infile.read().split("\n") if len(x.strip())]

    # Version prefix
    if "version" in data[0]:
        ret["version"] = data.pop(0).split()[1]

    # Box prefix
    box = _drop_prefix(data.pop(0).split(), 4)
    ret["box"] = tuple(float(x) for x in box)

    # Vertex prefix
    n_vertices = _drop_prefix(data.pop(0).split(), 2)
    n_vertices = int(n_vertices[0])
    vertices, data = data[:n_vertices], data[n_vertices:]
    ret["n_vertices"] = n_vertices
    ret["vertices"] = np.array([x.split() for x in vertices], dtype=np.float64)

    # Face prefix
    n_faces = _drop_prefix(data.pop(0).split(), 2)
    n_faces = int(n_faces[0])
    faces, data = data[:n_faces], data[n_faces:]
    ret["n_faces"] = n_faces
    ret["faces"] = np.array([x.split() for x in faces], dtype=np.float64)

    while len(data):
        if not data[0].startswith("inclusion"):
            data.pop(0)
        break

    if len(data) == 0:
        return ret

    n_inclusions = _drop_prefix(data.pop(0).split(), 2)
    n_inclusions = int(n_inclusions[0])
    incl, data = data[:n_inclusions], data[n_inclusions:]
    ret["n_inclusions"] = n_inclusions
    ret["inclusions"] = np.array([x.split() for x in incl], dtype=np.float64)

    return ret


def _read_vtu_file(file_path: str) -> Dict:
    """
    Parse a VTK XML file into a dictionary of numpy arrays.

    Parameters
    ----------
    file_path : str
        The path to the topology file to be parsed.

    Returns
    -------
    Dict
        Topology file content.
    """
    with open(file_path, mode="r") as ifile:
        data = ifile.read()

    root = ET.fromstring(data)
    piece = root.find(".//Piece")

    result = {
        "num_points": int(piece.get("NumberOfPoints")),
        "num_cells": int(piece.get("NumberOfCells")),
        "point_data": {},
        "points": None,
        "connectivity": None,
        "offsets": None,
        "types": None,
    }

    # Parse point data arrays
    if (point_data := piece.find("PointData")) is not None:
        for array in point_data.findall("DataArray"):
            data_type = _parse_dtype(array)
            result["point_data"][array.get("Name")] = _parse_data_array(
                array, data_type
            )

    if (points_array := piece.find(".//Points/DataArray")) is not None:
        data_type = _parse_dtype(points_array)
        result["points"] = _parse_data_array(points_array, data_type)

    if (cells := piece.find("Cells")) is not None:
        for array in cells.findall("DataArray"):
            data_type = _parse_dtype(array)
            result[array.get("Name")] = _parse_data_array(array, data_type)

    return result


def load_density(filename: str, **kwargs):
    """
    Load 3D density data from file.

    Parameters
    ----------
    filename : str
        Path to density file.
    **kwargs
        Additional keyword arguments passed to Density.from_file.

    Returns
    -------
    Density
        Loaded density object.
    """
    from tme import Density

    volume = Density.from_file(filename, **kwargs)

    if np.allclose(volume.sampling_rate, 0):
        warnings.warn(
            "All sampling rates are 0 - Setting them to 1 for now. Some functions might"
            "not behave properly. Make sure to define sampling rates if you forgot."
        )
        volume.sampling_rate = 1

    return volume



from importlib_resources import files

__all__ = [
    "Colors",
    "QGroupBox_style",
    "QPushButton_style",
    "QSpinBox_style",
    "QDoubleSpinBox_style",
    "QComboBox_style",
    "QCheckBox_style",
    "QLineEdit_style",
    "QScrollArea_style",
    "HelpLabel_style",
    "QTabBar_style",
    "QListWidget_style",
    "QSlider_style",
    "QMessageBox_style",
    "QProgressBar_style",
    "QToolButton_style",
    "QMenu_style",
    "QDockWidget_style",
    "QTable_style",
]


class Colors:
    """Centralized color definitions for the Mosaic application."""

    PRIMARY = "#4f46e5"  # Indigo - main accent color

    # =========================================================================
    # Text Colors
    # =========================================================================
    TEXT_PRIMARY = "#334155"  # Slate 700 - main text, headings
    TEXT_SECONDARY = "#64748b"  # Slate 500 - secondary text, labels, icons
    TEXT_MUTED = "#6b7280"  # Gray 500 - muted/placeholder text

    # =========================================================================
    # Border Colors
    # =========================================================================
    BORDER_HOVER = "#94a3b8"  # Slate 400 - border on hover
    BORDER_DARK = "#cbd5e1"  # Slate 300 - default borders

    # =========================================================================
    # Background Colors
    # =========================================================================
    BG_SECONDARY = "#f8fafc"  # Slate 50 - subtle background
    BG_TERTIARY = "#f1f5f9"  # Slate 100 - disabled/muted background
    BG_HOVER = "rgba(0, 0, 0, 0.06)"  # Hover state
    BG_PRESSED = "rgba(0, 0, 0, 0.10)"  # Pressed state

    # =========================================================================
    # Status Colors
    # =========================================================================
    SUCCESS = "#10b981"  # Emerald 500
    SUCCESS_BG = "#d1fae5"  # Emerald 100
    SUCCESS_TEXT = "#065f46"  # Emerald 800

    WARNING = "#f59e0b"  # Amber 500
    WARNING_DARK = "#d97706"  # Amber 600
    WARNING_BG = "#fef3c7"  # Amber 100
    WARNING_TEXT = "#92400e"  # Amber 800

    ERROR = "#ef4444"  # Red 500
    ERROR_BG = "#fee2e2"  # Red 100
    ERROR_TEXT = "#991b1b"  # Red 800

    NEUTRAL = "#6b7280"  # Gray 500
    NEUTRAL_BG = "#e5e7eb"  # Gray 200
    NEUTRAL_TEXT = "#374151"  # Gray 700

    # =========================================================================
    # Icon Colors
    # =========================================================================
    ICON = "#64748b"  # Slate 500 - primary icon color
    ICON_MUTED = "#9ca3af"  # Gray 400 - muted icons, chevrons

    WIDGET_HEIGHT = 30  # Standard height for input widgets (px)

    # =========================================================================
    # Entity Colors (for "By Entity" coloring mode)
    # =========================================================================
    ENTITY = [
        (0.18, 0.62, 0.78),  # Cerulean
        (0.98, 0.75, 0.18),  # Saffron
        (0.32, 0.70, 0.40),  # Malachite
        (0.72, 0.32, 0.78),  # Amethyst
        (0.95, 0.50, 0.20),  # Tangerine
        (0.22, 0.42, 0.72),  # Cobalt
        (0.85, 0.35, 0.55),  # Cerise
        (0.45, 0.75, 0.30),  # Chartreuse
        (0.55, 0.25, 0.60),  # Plum
        (0.20, 0.68, 0.58),  # Viridian
        (0.92, 0.58, 0.45),  # Coral
        (0.35, 0.35, 0.75),  # Ultramarine
        (0.75, 0.72, 0.25),  # Olive gold
        (0.78, 0.25, 0.38),  # Carmine
        (0.25, 0.78, 0.72),  # Turquoise
        (0.65, 0.45, 0.20),  # Bronze
        (0.52, 0.58, 0.85),  # Periwinkle
        (0.88, 0.40, 0.70),  # Fuchsia
        (0.38, 0.58, 0.28),  # Moss
        (0.70, 0.55, 0.65),  # Mauve
        (0.28, 0.55, 0.45),  # Teal
        (0.82, 0.65, 0.55),  # Peach
        (0.48, 0.38, 0.55),  # Grape
        (0.90, 0.25, 0.20),  # Vermillion red
    ]

    # Annotation colors — ENTITY palette brightened for visibility on dark backgrounds
    ANNOTATION = [tuple(min(1.0, c + 0.25) for c in color) for color in ENTITY]

    # =========================================================================
    # Category Colors (for pipelines, animations, etc.)
    # =========================================================================
    CATEGORY = {
        # Pipeline operation categories
        "input": "#7c3aed",  # Violet 600
        "preprocessing": "#2563eb",  # Blue 600
        "parametrization": "#059669",  # Emerald 600
        "analysis": "#0891b2",  # Cyan 600
        "export": "#ea580c",  # Orange 600
        # Animation types
        "trajectory": "#3b82f6",  # Blue 500
        "camera": "#10b981",  # Emerald 500
        "zoom": "#14b8a6",  # Teal 500
        "volume": "#f59e0b",  # Amber 500
        "visibility": "#8b5cf6",  # Violet 500
        "waypoint": "#ec4899",  # Pink 500
        # Pipeline presets
        "clear": "#ef4444",  # Red 500
        "import": "#3b82f6",  # Blue 500
        "cleanup": "#8b5cf6",  # Violet 500
        "meshing": "#10b981",  # Emerald 500
        "particle_picking": "#f59e0b",  # Amber 500
    }


def _get_resource_path(resource_name):
    """Get the absolute path to a resource in the package.

    Args:
        resource_name (str): Relative path to the resource within the
            package data directory

    Returns:
        str: The absolute path to the resource
    """
    return str(files("mosaic.data").joinpath(f"data/{resource_name}"))


HelpLabel_style = f"""
    QLabel {{
        color: {Colors.TEXT_MUTED};
        font-size: 12px;
        border-top: 0px;
    }}
"""

QGroupBox_style = f"""
    QGroupBox {{
        font-weight: 500;
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 6px;
        margin-top: 6px;
        padding-top: 14px;
    }}
    QGroupBox::title {{
        subcontrol-origin: margin;
        left: 7px;
        padding: 0px 5px 0px 5px;
    }}
"""

QPushButton_style = f"""
    QPushButton {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 6px 12px;
    }}
    QPushButton:hover {{
        background: {Colors.BG_HOVER};
        border: 1px solid rgba(0, 0, 0, 0.08);
    }}
    QPushButton:pressed {{
        background: {Colors.BG_PRESSED};
        border: 1px solid rgba(0, 0, 0, 0.12);
    }}
    QPushButton:focus {{
        outline: none;
    }}
"""

QLineEdit_style = f"""
    QLineEdit {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 4px 8px;
        selection-background-color: rgba(99, 102, 241, 0.6);
        background: transparent;
    }}
    QLineEdit:focus {{
        outline: none;
        border: 1px solid {Colors.PRIMARY};
    }}
    QLineEdit:hover:!focus {{
        border: 1px solid {Colors.BORDER_HOVER};
    }}
    QLineEdit:disabled {{
        background-color: {Colors.BG_TERTIARY};
        color: {Colors.BORDER_HOVER};
    }}
"""

QSpinBox_style = f"""
    QSpinBox {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 4px 8px;
        background-color: transparent;
        selection-background-color: rgba(99, 102, 241, 0.6);
    }}
    QSpinBox:focus {{
        outline: none;
        border: 1px solid {Colors.PRIMARY};
    }}
    QSpinBox:hover:!focus {{
        border: 1px solid {Colors.BORDER_HOVER};
    }}
    QSpinBox:disabled {{
        background-color: {Colors.BG_TERTIARY};
        color: {Colors.BORDER_HOVER};
    }}
    QSpinBox::up-button, QSpinBox::down-button {{
        border: 1px solid {Colors.BORDER_DARK};
        width: 16px;
        background-color: {Colors.BG_SECONDARY};
    }}
    QSpinBox::up-button {{
        border-top-right-radius: 3px;
        border-bottom: none;
    }}
    QSpinBox::down-button {{
        border-bottom-right-radius: 3px;
    }}
    QSpinBox::up-button:hover, QSpinBox::down-button:hover {{
        background-color: {Colors.BG_TERTIARY};
        border-color: {Colors.BORDER_HOVER};
    }}
    QSpinBox::up-button:pressed, QSpinBox::down-button:pressed {{
        background-color: #e2e8f0;
    }}
"""


QDoubleSpinBox_style = f"""
    QDoubleSpinBox {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 4px 8px;
        background-color: transparent;
        selection-background-color: rgba(99, 102, 241, 0.6);
    }}
    QDoubleSpinBox:focus {{
        outline: none;
        border: 1px solid {Colors.PRIMARY};
    }}
    QDoubleSpinBox:hover:!focus {{
        border: 1px solid {Colors.BORDER_HOVER};
    }}
    QDoubleSpinBox:disabled {{
        background-color: {Colors.BG_TERTIARY};
        color: {Colors.BORDER_HOVER};
    }}
    QDoubleSpinBox::up-button, QDoubleSpinBox::down-button {{
        border: 1px solid {Colors.BORDER_DARK};
        width: 16px;
        background-color: {Colors.BG_SECONDARY};
    }}
    QDoubleSpinBox::up-button {{
        border-top-right-radius: 3px;
        border-bottom: none;
    }}
    QDoubleSpinBox::down-button {{
        border-bottom-right-radius: 3px;
    }}
    QDoubleSpinBox::up-button:hover, QDoubleSpinBox::down-button:hover {{
        background-color: {Colors.BG_TERTIARY};
        border-color: {Colors.BORDER_HOVER};
    }}
    QDoubleSpinBox::up-button:pressed, QDoubleSpinBox::down-button:pressed {{
        background-color: #e2e8f0;
    }}
"""


QComboBox_style = f"""
    QComboBox {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        min-height: 27px;
        padding: 0px 8px;
        background: transparent;
        selection-background-color: rgba(99, 102, 241, 0.6);
    }}
    QComboBox:focus {{
        outline: none;
        border: 1px solid {Colors.PRIMARY};
    }}
    QComboBox:hover:!focus {{
        border: 1px solid {Colors.BORDER_HOVER};
    }}
    QComboBox:disabled {{
        background-color: {Colors.BG_TERTIARY};
        color: {Colors.BORDER_HOVER};
    }}
    QComboBox::drop-down:disabled {{
        border: none;
    }}
    QComboBox QAbstractItemView {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        selection-background-color: rgba(99, 102, 241, 0.3);
    }}
"""

QCheckBox_style = f"""
    QCheckBox {{
        spacing: 5px;
        background-color: transparent;
    }}
    QCheckBox::indicator {{
        width: 18px;
        height: 18px;
        border: 1px solid {Colors.BORDER_DARK};
    }}
    QCheckBox::indicator:hover {{
        border: 1px solid {Colors.BORDER_DARK};;
    }}
    QCheckBox::indicator:focus {{
        border: 1px solid {Colors.BORDER_DARK};
    }}
    QCheckBox::indicator:checked {{
        image: url('{_get_resource_path("checkbox-checkmark.svg")}')
    }}
"""

QScrollArea_style = f"""
    QScrollArea {{
        border: none;
    }}
    QScrollBar:vertical {{
        border: none;
        background: {Colors.BG_TERTIARY};
        width: 6px;
        margin: 0px;
    }}
    QScrollBar::handle:vertical {{
        background: {Colors.BORDER_DARK};
        min-height: 20px;
        border-radius: 3px;
    }}
    QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {{
        height: 0px;
    }}
    QScrollBar:horizontal {{
        height: 0px;
        background: transparent;
    }}
"""

QTabBar_style = f"""
    QTabBar::tab {{
        background: transparent;
        border: 1px solid {Colors.BORDER_DARK};
        border-bottom: none;
        border-top-left-radius: 6px;
        border-top-right-radius: 6px;
        padding: 6px 12px;
        margin-right: 2px;
    }}
    QTabBar::tab:selected {{
        color: {Colors.PRIMARY};
        border-color: {Colors.PRIMARY};
    }}
    QTabBar::tab:hover:!selected {{
        color: {Colors.TEXT_MUTED};
    }}
    /* Style for the tab widget itself */
    QTabWidget::pane {{
        border: 1px solid {Colors.BORDER_DARK};
        background-color: transparent;
        border-top-right-radius: 6px;
        border-bottom-right-radius: 6px;
        border-bottom-left-radius: 6px;
    }}

    /* Style for the tab contents */
    QWidget#scrollContentWidget {{
        background-color: transparent;
    }}

    /* Make scroll areas transparent */
    QScrollArea {{
        background-color: transparent;
        border: none;
    }}

"""


QTable_style = f"""
    QTableWidget {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        background-color: transparent;
        outline: none;
        gridline-color: {Colors.BORDER_DARK};
    }}
    QTableWidget::item {{
        border: none;
    }}
    QTableWidget::item:hover {{
        background-color: {Colors.BG_HOVER};
    }}
    QTableWidget::item:selected {{
        background-color: rgba(99, 102, 241, 0.3);
    }}
    QTableWidget QHeaderView::section {{
        background: transparent;
        border: none;
        border-bottom: 1px solid {Colors.BORDER_DARK};
        padding: 4px 4px;
        color: {Colors.TEXT_SECONDARY};
    }}
    QTableWidget QTableCornerButton::section {{
        background: transparent;
        border: none;
        border-bottom: 1px solid {Colors.BORDER_DARK};
    }}
"""

QListWidget_style = f"""
    QListWidget {{
        border: none;
        background-color: transparent;
        outline: none;
        padding: 4px 0px;
    }}
    QListWidget::item {{
        border-radius: 6px;
        margin: 2px 8px;
        font-size: 13px;
    }}
    QListWidget::item:hover {{
        background-color: {Colors.BG_PRESSED};
    }}
    QListWidget::item:selected {{
        background-color: rgba(99, 102, 241, 0.3);
        font-weight: 500;
    }}
"""

# Left background used to be #4f46e5
QSlider_style = f"""
    QSlider {{
        height: 24px;
    }}
    QSlider:disabled {{
        opacity: 0.5;
    }}
    QSlider::groove:horizontal {{
        height: 4px;
        background: #e2e8f0;
        border-radius: 2px;
    }}
    QSlider::groove:horizontal:disabled {{
        background: {Colors.BG_TERTIARY};
    }}
    QSlider::handle:horizontal {{
        background: #ffffff;
        border: 1px solid {Colors.BORDER_DARK};
        width: 16px;
        height: 16px;
        margin: -6px 0;
        border-radius: 8px;
    }}
    QSlider::handle:horizontal:hover {{
        border-color: {Colors.PRIMARY};
    }}
    QSlider::handle:horizontal:focus {{
        border: 1px solid {Colors.PRIMARY};
        background: #f9fafb;
    }}
    QSlider::handle:horizontal:disabled {{
        background: {Colors.BG_SECONDARY};
        border: 1px solid #e2e8f0;
    }}
    QSlider::sub-page:horizontal {{
        background: {Colors.BORDER_HOVER};
        border-radius: 2px;
    }}
    QSlider::sub-page:horizontal:disabled {{
        background: {Colors.BORDER_DARK};
    }}
"""

QMessageBox_style = f"""
    QMessageBox QLabel {{
        font-size: 13px;
    }}
    QMessageBox QPushButton {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 6px 16px;
        min-width: 80px;
    }}
    QMessageBox QPushButton:hover {{
        border: 1px solid {Colors.ICON_MUTED};
        background: {Colors.BG_HOVER};
    }}
    QMessageBox QPushButton:pressed {{
        border: 1px solid {Colors.ICON_MUTED};
        background: rgba(0, 0, 0, 0.24);
    }}
    QMessageBox QPushButton:focus {{
        outline: none;
    }}
    QMessageBox QCheckBox {{
        color: {Colors.TEXT_PRIMARY};
        font-size: 12px;
    }}
    QMessageBox QTextEdit {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 4px;
        padding: 8px;
    }}
"""

QProgressBar_style = f"""
    QProgressBar {{
        border: none;
        background-color: {Colors.NEUTRAL_BG};
        border-radius: 4px;
        height: 8px;
    }}
    QProgressBar::chunk {{
        background-color: {Colors.PRIMARY};
        border-radius: 4px;
    }}
"""


QToolButton_style = f"""
    QToolButton {{
        min-width: 52px;
        padding: 4px 6px;
        border-radius: 6px;
        font-size: 11px;
        background: transparent;
        border: 1px solid transparent;
    }}
    QToolButton:hover {{
        background: {Colors.BG_HOVER};
        border: 1px solid rgba(0, 0, 0, 0.08);
    }}
    QToolButton:pressed {{
        background: {Colors.BG_PRESSED};
        border: 1px solid rgba(0, 0, 0, 0.12);
    }}
    QToolButton::menu-indicator {{
        image: url(none);
        width: 0px;
        subcontrol-position: right bottom;
        subcontrol-origin: padding;
        margin-left: 0px;
    }}
    QToolButton::menu-button {{
        border: 1px solid transparent;
        width: 14px;
        padding: 0px;
        margin: 0px;
        border-radius: 4px;
    }}
    QToolButton::menu-button:hover {{
        background: {Colors.BG_HOVER};
    }}
"""

QMenu_style = f"""
    QMenu {{
        border: 1px solid {Colors.BORDER_DARK};
        border-radius: 8px;
        padding: 4px;
    }}
    QMenu::item {{
        padding: 4px 12px;
        border-radius: 4px;
        border: 1px solid transparent;
    }}
    QMenu::item:selected {{
        background: {Colors.BG_HOVER};
        border: 1px solid rgba(0, 0, 0, 0.08);
    }}
    QMenu::item:pressed {{
        background: {Colors.BG_PRESSED};
        border: 1px solid rgba(0, 0, 0, 0.12);
    }}
    QMenu::separator {{
        height: 1px;
        background-color: {Colors.BG_PRESSED};
        margin: 4px 8px;
    }}
    QMenu::indicator {{
        width: 16px;
        height: 16px;
        margin-right: 6px;
    }}
    QMenu::indicator:checked {{
        image: url('{_get_resource_path("checkbox-checkmark.svg")}');
    }}
"""

QDockWidget_style = f"""
    QDockWidget {{
        titlebar-close-icon: url('{_get_resource_path("dock-close.svg")}');
        titlebar-normal-icon: url('{_get_resource_path("dock-float.svg")}');
    }}
    QDockWidget::title {{
        background: transparent;
        text-align: left;
        padding-top: 2px;
    }}
    QDockWidget::close-button, QDockWidget::float-button {{
        border: 1px solid {Colors.BORDER_DARK};
        background: transparent;
        width: 24px;
        height: 24px;
        max-width: 24px;
        max-height: 24px;
        border-radius: 4px;
        subcontrol-origin: padding;
    }}
    QDockWidget::close-button {{
        subcontrol-position: right center;
        right: 4px;
    }}
    QDockWidget::float-button {{
        subcontrol-position: right center;
        right: 24px;
    }}
    QDockWidget::close-button:hover, QDockWidget::float-button:hover {{
        background: {Colors.BG_HOVER};
        border: 1px solid rgba(0, 0, 0, 0.08);
    }}
    QDockWidget::close-button:pressed, QDockWidget::float-button:pressed {{
        background: {Colors.BG_PRESSED};
        border: 1px solid rgba(0, 0, 0, 0.12);
    }}
"""



"""
Animation composer dialog for creating timeline-based animations.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import json
from dataclasses import dataclass
from typing import Dict, List, Any
from uuid import uuid4

import imageio
from qtpy.QtCore import Qt, QTimer, QSize
from qtpy.QtGui import QKeySequence, QShortcut
from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QGridLayout,
    QPushButton,
    QToolButton,
    QSpinBox,
    QGroupBox,
    QFileDialog,
    QMessageBox,
    QApplication,
    QSizePolicy,
    QCheckBox,
)
import qtawesome as qta

from .timeline import TimelineWidget
from .animations import AnimationType, BaseAnimation
from .settings import AnimationSettings, ExportDialog
from ._utils import FrameWriter, capture_frame

from ..__version__ import __version__
from ..utils import Throttle
from ..stylesheets import (
    QMessageBox_style,
    QLineEdit_style,
    QSpinBox_style,
    QDoubleSpinBox_style,
    QComboBox_style,
    QCheckBox_style,
    QSlider_style,
    QGroupBox_style,
    QListWidget_style,
    QPushButton_style,
    QScrollArea_style,
    Colors,
)


@dataclass
class Track:
    id: str
    animation: BaseAnimation
    color: str


class AnimationComposerDialog(QDialog):
    """Dialog for composing complex, timeline-based animations."""

    FORMAT_SETTINGS = {
        "MP4": {"ext": ".mp4", "video": True},
        "AVI": {"ext": ".avi", "video": True},
        "PNG Sequence": {"ext": ".png", "video": False},
    }

    def __init__(self, vtk_widget, volume_viewer=None, cdata=None, parent=None):
        super().__init__(parent)
        self.tracks: List[Track] = []
        self.selected_track = None
        self.current_frame = 0

        self.is_playing = False
        self.is_looping = False
        self.playback_fps = 30
        self.timer = QTimer()
        self.timer.timeout.connect(self._advance_frame)

        self.cdata = cdata
        self.vtk_widget = vtk_widget
        self.volume_viewer = volume_viewer

        self._setup_ui()
        self._setup_shortcuts()

        self.setStyleSheet(
            QMessageBox_style
            + QLineEdit_style
            + QSpinBox_style
            + QDoubleSpinBox_style
            + QComboBox_style
            + QCheckBox_style
            + QSlider_style
            + QGroupBox_style
            + QListWidget_style
            + QPushButton_style
            + QScrollArea_style
        )

    def _setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(8, 8, 8, 8)
        main_layout.setSpacing(8)

        # Icon toolbar for project actions
        toolbar = QWidget()
        toolbar_layout = QHBoxLayout(toolbar)
        toolbar_layout.setContentsMargins(0, 0, 0, 0)
        toolbar_layout.setSpacing(4)

        for icon, tooltip, callback in [
            ("ph.folder-open", "Load project", self.load_project),
            ("ph.floppy-disk", "Save project", self.save_project),
            ("ph.trash", "Clear all tracks", self.clear_all_tracks),
            ("ph.export", "Export animation", self.export_animation),
        ]:
            btn = QToolButton()
            btn.setIcon(qta.icon(icon, color=Colors.ICON))
            btn.setToolTip(tooltip)
            btn.setIconSize(QSize(20, 20))
            btn.clicked.connect(callback)
            toolbar_layout.addWidget(btn)

        toolbar_layout.addStretch()

        # Loop checkbox
        self.loop_checkbox = QCheckBox("Loop")
        self.loop_checkbox.setToolTip("Loop playback (L)")
        self.loop_checkbox.stateChanged.connect(self._on_loop_changed)
        toolbar_layout.addWidget(self.loop_checkbox)

        main_layout.addWidget(toolbar)

        # Playback controls
        controls = QWidget()
        controls_layout = QHBoxLayout(controls)
        controls_layout.setContentsMargins(0, 0, 0, 0)
        controls_layout.setSpacing(4)

        back_btn = QToolButton()
        back_btn.setIcon(qta.icon("ph.skip-back"))
        back_btn.setToolTip("Go to start (Home)")
        back_btn.clicked.connect(lambda: self.set_current_frame(0))

        self.play_btn = QToolButton()
        self.play_btn.setIcon(qta.icon("ph.play"))
        self.play_btn.setToolTip("Play/Pause (Space)")
        self.play_btn.clicked.connect(self.toggle_play)

        forward_btn = QToolButton()
        forward_btn.setIcon(qta.icon("ph.skip-forward"))
        forward_btn.setToolTip("Go to end (End)")
        forward_btn.clicked.connect(self._go_to_end)

        controls_layout.addWidget(back_btn)
        controls_layout.addWidget(self.play_btn)
        controls_layout.addWidget(forward_btn)

        self._frame_throttle = Throttle(self.set_current_frame, interval_ms=50)

        self.frame_spin = QSpinBox()
        self.frame_spin.setRange(0, 2 << 29)
        self.frame_spin.valueChanged.connect(self._frame_throttle)
        controls_layout.addWidget(self.frame_spin, 1)

        main_layout.addWidget(controls)

        # Timeline
        self.timeline = TimelineWidget()
        self.timeline.content.trackSelected.connect(self._on_track_selected)
        self.timeline.frameMoved.connect(self._frame_throttle)
        self.timeline.trackMoved.connect(self._on_track_moved)
        self.timeline.trackRemoved.connect(self.delete_track)
        self.timeline.setMinimumHeight(120)
        self.timeline.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        main_layout.addWidget(self.timeline, 1)

        # Presets group
        presets_group = QGroupBox("Presets")
        presets_layout = QVBoxLayout(presets_group)
        presets_layout.setContentsMargins(8, 12, 8, 8)
        presets_layout.setSpacing(6)

        presets_row = QHBoxLayout()
        presets_row.setSpacing(6)
        for name, tooltip, preset_type in [
            ("Trajectory", "Animate through trajectory frames", "trajectory"),
            ("Volume", "Animate through volume slices", "slices"),
            (
                "Flythrough",
                "Slice through volume, reveal actors, then slice back",
                "reveal_flythrough",
            ),
        ]:
            btn = QPushButton(name)
            btn.setToolTip(tooltip)
            btn.clicked.connect(lambda _, p=preset_type: self._load_preset(p))
            presets_row.addWidget(btn)
        presets_layout.addLayout(presets_row)
        main_layout.addWidget(presets_group)

        # Add Animation group (2 rows x 3 cols for compactness)
        anims_group = QGroupBox("Add Animation")
        anims_layout = QGridLayout(anims_group)
        anims_layout.setContentsMargins(8, 12, 8, 8)
        anims_layout.setSpacing(6)

        anim_types = list(AnimationType)
        cols = 3
        for i, anim_type in enumerate(anim_types):
            btn = QPushButton(anim_type.value["name"])
            btn.setToolTip(anim_type.value.get("description", ""))
            btn.clicked.connect(lambda _, t=anim_type: self.add_animation(t))
            anims_layout.addWidget(btn, i // cols, i % cols)
        main_layout.addWidget(anims_group)

        # Track settings
        self.properties_panel = AnimationSettings()
        self.properties_panel.animationChanged.connect(self._on_animation_changed)
        main_layout.addWidget(self.properties_panel)

        self.setMinimumWidth(280)
        self.resize(300, 700)

    def _load_preset(self, preset_type: str):
        """Load a preset animation configuration."""
        self._cleanup_tracks()
        self.tracks.clear()
        self.timeline.set_tracks(self.tracks)

        if preset_type == "trajectory":
            self._load_trajectory_preset()
        elif preset_type == "slices":
            self._load_slices_preset()
        elif preset_type == "reveal_flythrough":
            self._load_reveal_flythrough_preset()

        if self.tracks:
            self._on_track_selected(self.tracks[0].id)

    def _load_trajectory_preset(self):
        """Load preset for trajectory animation."""
        from mosaic.geometry import GeometryTrajectory

        trajectories = [
            x for x in self.cdata._models.data if isinstance(x, GeometryTrajectory)
        ]

        if not trajectories:
            QMessageBox.warning(
                self,
                "No Trajectories",
                "No trajectory data found. Load a trajectory first.",
            )
            return

        self.add_animation(AnimationType.TRAJECTORY)

    def _load_slices_preset(self):
        """Load preset for volume slice animation."""
        if self.volume_viewer is None:
            QMessageBox.warning(
                self, "No Volume", "No volume viewer available for slice animation."
            )
            return

        volume = getattr(self.volume_viewer.primary, "volume", None)
        if volume is None:
            QMessageBox.warning(
                self, "No Volume", "No volume loaded. Load a volume first."
            )
            return

        self.add_animation(AnimationType.SLICE)

    def _load_reveal_flythrough_preset(self):
        """Load preset for reveal flythrough animation.

        This creates a sequence that:
        1. Slices through the volume (forward) with all actors hidden
        2. Reveals all actors at the midpoint
        3. Slices back through the volume (backward)
        """
        if self.volume_viewer is None:
            QMessageBox.warning(
                self, "No Volume", "No volume viewer available for reveal flythrough."
            )
            return

        volume = getattr(self.volume_viewer.primary, "volume", None)
        if volume is None:
            QMessageBox.warning(
                self, "No Volume", "No volume loaded. Load a volume first."
            )
            return

        # Get all actor object IDs
        all_objects = []
        for _, obj in self.cdata.format_datalist("data"):
            all_objects.append(id(obj))
        for _, obj in self.cdata.format_datalist("models"):
            all_objects.append(id(obj))

        # 1. Hide all actors at the start (instant opacity to 0)
        hide_anim = self.add_animation(AnimationType.VISIBILITY)
        if hide_anim:
            hide_anim.name = "Hide Actors"
            hide_anim.update_parameters(
                start_opacity=1.0,
                target_opacity=0.0,
                easing="instant",
                selected_objects=all_objects,
            )
            hide_anim.start_frame = 0
            hide_anim.stop_frame = 1

        # 2. Volume slice forward
        forward_slice = self.add_animation(AnimationType.SLICE)
        if forward_slice:
            forward_slice.name = "Slice Forward"
            forward_slice.update_parameters(direction="forward")

        # 3. Reveal all actors at midpoint (instant opacity to 1)
        reveal_anim = self.add_animation(AnimationType.VISIBILITY)
        if reveal_anim:
            reveal_anim.name = "Reveal Actors"
            reveal_anim.update_parameters(
                start_opacity=0.0,
                target_opacity=1.0,
                easing="instant",
                selected_objects=all_objects,
            )
            reveal_anim.start_frame = 0
            reveal_anim.stop_frame = 1

        # 4. Volume slice backward
        backward_slice = self.add_animation(AnimationType.SLICE)
        if backward_slice:
            backward_slice.name = "Slice Backward"
            backward_slice.update_parameters(direction="backward")

        # Update timeline
        self.timeline.set_tracks(self.tracks)

    def add_animation(self, anim_type: AnimationType):
        """Add a new animation track."""
        animation_class = anim_type.value["class"]

        try:
            animation = animation_class(
                cdata=self.cdata,
                vtk_widget=self.vtk_widget,
                volume_viewer=self.volume_viewer,
                global_start_frame=self._get_next_start_frame(),
                enabled=True,
                name=f"{anim_type.value['name']} {len(self.tracks) + 1}",
            )
        except Exception as e:
            QMessageBox.warning(self, "Error", f"Failed to create animation: {e}")
            return None

        track = Track(
            id=str(uuid4()),
            animation=animation,
            color=anim_type.value["color"],
        )

        self.tracks.append(track)
        self.timeline.set_tracks(self.tracks)

        return animation

    def _get_next_start_frame(self) -> int:
        """Calculate the next available start frame for a new track."""
        if not self.tracks:
            return 0
        return max(
            t.animation.global_start_frame + t.animation.duration for t in self.tracks
        )

    def _cleanup_tracks(self, tracks=None):
        """Reset animation state and remove preview actors for the given tracks."""
        for track in tracks or self.tracks:
            track.animation.reset()
            if hasattr(track.animation, "cleanup_preview"):
                track.animation.cleanup_preview()

    def _get_track(self, track_id: str):
        return next((t for t in self.tracks if t.id == track_id), None)

    def delete_track(self, track_id: str):
        """Remove a track from the timeline."""
        removed = [t for t in self.tracks if t.id == track_id]
        self._cleanup_tracks(removed)
        self.tracks = [t for t in self.tracks if t.id != track_id]
        if self.selected_track == track_id:
            self.selected_track = None
        self.timeline.set_tracks(self.tracks)

    def _on_track_selected(self, track_id: str):
        if (track := self._get_track(track_id)) is None:
            return

        self.selected_track = track_id
        self.properties_panel.set_animation(track.animation)

    def _on_track_moved(self, track_id: str, new_frame: int):
        if (track := self._get_track(track_id)) is None:
            return

        track.animation.global_start_frame = new_frame
        self.properties_panel.global_start_spin.setValue(new_frame)
        self.timeline.update()

    def _on_animation_changed(self, changes: Dict[str, Any]):
        self.timeline.update()

    def _get_total_frames(self) -> int:
        """Get the total number of frames across all tracks."""
        if not self.tracks:
            return 0
        return max(
            t.animation.global_start_frame + t.animation.duration for t in self.tracks
        )

    def set_current_frame(self, frame: int):
        """Set the current playback frame and update all animations."""
        if not self.tracks:
            return

        total_frames = self._get_total_frames()
        self.current_frame = max(0, min(total_frames, frame))

        # Prevent recursive updates from spinbox
        self.frame_spin.blockSignals(True)
        self.frame_spin.setValue(self.current_frame)
        self.frame_spin.blockSignals(False)

        self.timeline.set_current_frame(self.current_frame)

        # Stop playback at end (unless looping, which is handled in _advance_frame)
        if (
            self.current_frame >= total_frames
            and self.is_playing
            and not self.is_looping
        ):
            self.toggle_play()

        for track in self.tracks:
            track.animation.reset()
        for track in self.tracks:
            track.animation.update(self.current_frame)

        self.vtk_widget.GetRenderWindow().Render()

    def _go_to_end(self):
        """Jump to the last frame."""
        self.set_current_frame(self._get_total_frames())

    def _set_preview_visibility(self, visible: bool):
        """Show or hide preview actors (e.g. waypoint path) on all tracks."""
        for track in self.tracks:
            for actor in getattr(track.animation, "_path_actors", []):
                actor.SetVisibility(visible)
        self.vtk_widget.GetRenderWindow().Render()

    def toggle_play(self):
        """Toggle playback state."""
        self.is_playing = not self.is_playing

        if self.is_playing:
            self._set_preview_visibility(False)
            self.play_btn.setIcon(qta.icon("ph.pause"))
            self.timer.start(1000 // self.playback_fps)
        else:
            self._set_preview_visibility(True)
            self.play_btn.setIcon(qta.icon("ph.play"))
            self.timer.stop()

    def _advance_frame(self):
        if not self.is_playing:
            return

        next_frame = self.current_frame + 1
        total_frames = self._get_total_frames()

        if next_frame >= total_frames:
            if self.is_looping:
                next_frame = 0
            else:
                self.set_current_frame(total_frames)
                return

        self.set_current_frame(next_frame)

    def _on_loop_changed(self, state):
        """Handle loop checkbox state change."""
        self.is_looping = state == Qt.CheckState.Checked.value

    def _setup_shortcuts(self):
        """Setup keyboard shortcuts."""
        # Spacebar for play/pause
        play_shortcut = QShortcut(QKeySequence(Qt.Key.Key_Space), self)
        play_shortcut.activated.connect(self.toggle_play)

        # L for loop toggle
        loop_shortcut = QShortcut(QKeySequence(Qt.Key.Key_L), self)
        loop_shortcut.activated.connect(self._toggle_loop)

        # Home for go to start
        home_shortcut = QShortcut(QKeySequence(Qt.Key.Key_Home), self)
        home_shortcut.activated.connect(lambda: self.set_current_frame(0))

        # End for go to end
        end_shortcut = QShortcut(QKeySequence(Qt.Key.Key_End), self)
        end_shortcut.activated.connect(self._go_to_end)

        # Left/Right arrow for frame stepping
        left_shortcut = QShortcut(QKeySequence(Qt.Key.Key_Left), self)
        left_shortcut.activated.connect(
            lambda: self.set_current_frame(self.current_frame - 1)
        )

        right_shortcut = QShortcut(QKeySequence(Qt.Key.Key_Right), self)
        right_shortcut.activated.connect(
            lambda: self.set_current_frame(self.current_frame + 1)
        )

    def closeEvent(self, event):
        """Clean up all tracks when the dialog is closed."""
        self._cleanup_tracks()
        super().closeEvent(event)

    def _toggle_loop(self):
        """Toggle loop playback."""
        self.loop_checkbox.setChecked(not self.loop_checkbox.isChecked())

    def clear_all_tracks(self):
        """Clear all animation tracks."""
        if not self.tracks:
            return

        reply = QMessageBox.question(
            self,
            "Clear All Tracks",
            "Are you sure you want to remove all animation tracks?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )

        if reply == QMessageBox.StandardButton.Yes:
            self._cleanup_tracks()
            self.tracks.clear()
            self.selected_track = None
            self.timeline.set_tracks(self.tracks)

    def _capture_frame(self, transparent_bg: bool = False, magnification: int = 1):
        """Capture the current frame from the VTK renderer."""
        render_window = self.vtk_widget.GetRenderWindow()
        return capture_frame(
            render_window, transparent_bg=transparent_bg, magnification=magnification
        )

    def export_animation(self):
        """Export the animation with current settings."""
        if not self.tracks:
            QMessageBox.warning(
                self, "No Animation", "Add at least one animation track before export."
            )
            return

        # Get current render window dimensions
        render_window = self.vtk_widget.GetRenderWindow()
        if render_window is None:
            QMessageBox.warning(self, "Error", "No render window available.")
            return

        current_size = render_window.GetSize()

        # Pause playback during export
        if self.is_playing:
            self.toggle_play()

        # Show export dialog
        total_frames = self._get_total_frames()
        dialog = ExportDialog(
            total_frames=total_frames,
            current_width=current_size[0],
            current_height=current_size[1],
            parent=self,
        )
        if not dialog.exec():
            return

        settings = dialog.get_settings()
        format_name = settings["format"]
        format_settings = self.FORMAT_SETTINGS.get(format_name, {})
        ext = format_settings.get("ext", ".mp4")
        is_video = format_settings.get("video", True)

        filename, _ = QFileDialog.getSaveFileName(
            self, "Save Animation", "", f"Animation (*{ext})"
        )

        if not filename:
            return

        if not filename.endswith(ext):
            filename += ext

        start_frame = settings["start_frame"]
        end_frame = settings["end_frame"]
        stride = settings["stride"]
        fps = settings["fps"]
        quality = settings["quality"]
        target_width = settings["width"]
        target_height = settings["height"]
        magnification = settings["magnification"]
        multisamples = settings["multisamples"]

        if start_frame >= end_frame:
            QMessageBox.warning(
                self, "Invalid Range", "Start frame must be less than end frame."
            )
            return

        frames = list(range(start_frame, end_frame + 1, stride))

        # Store original state
        original_frame = self.current_frame
        original_size = render_window.GetSize()

        # Use target dimensions from dialog, ensure even for video encoding
        width = target_width
        height = target_height
        if is_video:
            height += height % 2
            width += width % 2

        # Setup writer
        if is_video:
            quality_val = max(min(quality / 10.0, 10), 1)
            writer = imageio.get_writer(
                filename, mode="I", fps=fps, quality=quality_val, macro_block_size=None
            )
        else:
            writer = FrameWriter(filename)

        try:
            from mosaic.dialogs import ProgressDialog

            self._set_preview_visibility(False)
            render_window.SetOffScreenRendering(1)

            # Store original multisamples and apply export settings
            original_multisamples = render_window.GetMultiSamples()
            render_window.SetMultiSamples(multisamples)

            # Set render window to target dimensions
            # capture_frame will handle magnification (render larger, then downscale)
            render_window.SetSize(width, height)
            render_window.Render()

            with ProgressDialog(frames, title="Exporting Animation") as progress:
                for frame_idx in progress:
                    self.set_current_frame(frame_idx)
                    render_window.Render()

                    frame = self._capture_frame(
                        transparent_bg=not is_video, magnification=magnification
                    )
                    writer.append_data(frame)
                    QApplication.processEvents()

            writer.close()
            render_window.SetMultiSamples(original_multisamples)

            QMessageBox.information(
                self, "Export Complete", f"Animation saved to:\n{filename}"
            )

        except Exception as e:
            QMessageBox.warning(self, "Export Error", f"Failed to export: {e}")

        finally:
            self._set_preview_visibility(True)
            render_window.SetOffScreenRendering(0)
            render_window.SetSize(*original_size)
            render_window.Render()
            self.set_current_frame(original_frame)
            # Note: Don't restore playback state - export is a deliberate user action

    def save_project(self):
        """Save the animation project to a JSON file."""
        if not self.tracks:
            QMessageBox.warning(
                self, "No Animation", "Add at least one animation track before saving."
            )
            return

        filename, _ = QFileDialog.getSaveFileName(
            self, "Save Animation Project", "", "Animation Project (*.manim)"
        )

        if not filename:
            return

        if not filename.endswith(".manim"):
            filename += ".manim"

        project_data = {
            "version": __version__,
            "tracks": [],
        }

        for track in self.tracks:
            anim = track.animation
            anim_type = None
            for at in AnimationType:
                if isinstance(anim, at.value["class"]):
                    anim_type = at.name
                    break

            track_data = {
                "type": anim_type,
                "name": anim.name,
                "enabled": anim.enabled,
                "global_start_frame": anim.global_start_frame,
                "start_frame": anim.start_frame,
                "stop_frame": anim.stop_frame,
                "stride": anim.stride,
                "parameters": self._serialize_parameters(anim.parameters),
            }
            project_data["tracks"].append(track_data)

        try:
            with open(filename, "w") as f:
                json.dump(project_data, f, indent=2)

            QMessageBox.information(
                self, "Project Saved", f"Animation project saved to:\n{filename}"
            )
        except Exception as e:
            QMessageBox.warning(self, "Save Error", f"Failed to save project: {e}")

    def _serialize_parameters(self, params: Dict) -> Dict:
        """Convert parameters to JSON-serializable format."""
        serialized = {}
        for key, value in params.items():
            if isinstance(value, (list, tuple)):
                items = []
                for v in value:
                    if isinstance(v, dict):
                        items.append(
                            {
                                k: list(val) if isinstance(val, tuple) else val
                                for k, val in v.items()
                            }
                        )
                    elif isinstance(v, (int, float)):
                        items.append(float(v))
                    else:
                        items.append(str(v))
                serialized[key] = items
            elif isinstance(value, (int, float, str, bool)):
                serialized[key] = value
            else:
                serialized[key] = str(value)
        return serialized

    def load_project(self):
        """Load an animation project from a JSON file."""
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load Animation Project", "", "Animation Project (*.manim)"
        )

        if not filename:
            return

        # Pause playback during load
        if self.is_playing:
            self.toggle_play()

        try:
            with open(filename, "r") as f:
                project_data = json.load(f)
        except Exception as e:
            QMessageBox.warning(self, "Load Error", f"Failed to read project file: {e}")
            return

        self._cleanup_tracks()
        self.tracks.clear()
        self.timeline.set_tracks(self.tracks)
        for track_data in project_data.get("tracks", []):
            anim_type_name = track_data.get("type")
            if not anim_type_name:
                continue

            try:
                anim_type = AnimationType[anim_type_name]
            except KeyError:
                QMessageBox.warning(
                    self,
                    "Unknown Animation",
                    f"Unknown animation type: {anim_type_name}",
                )
                continue

            animation_class = anim_type.value["class"]

            try:
                animation = animation_class(
                    cdata=self.cdata,
                    vtk_widget=self.vtk_widget,
                    volume_viewer=self.volume_viewer,
                    global_start_frame=track_data.get("global_start_frame", 0),
                    enabled=track_data.get("enabled", True),
                    name=track_data.get("name", anim_type.value["name"]),
                )

                animation.start_frame = track_data.get("start_frame", 0)
                animation.stop_frame = track_data.get("stop_frame", 100)
                animation.stride = track_data.get("stride", 1)

                # Restore parameters
                for key, value in track_data.get("parameters", {}).items():
                    animation.update_parameters(**{key: value})

                track = Track(
                    id=str(uuid4()),
                    animation=animation,
                    color=anim_type.value["color"],
                )
                self.tracks.append(track)

            except Exception as e:
                QMessageBox.warning(
                    self, "Load Error", f"Failed to create animation: {e}"
                )
                continue

        self.timeline.set_tracks(self.tracks)

        if self.tracks:
            self._on_track_selected(self.tracks[0].id)

        QMessageBox.information(
            self,
            "Project Loaded",
            f"Loaded {len(self.tracks)} animation track(s) from:\n{filename}",
        )



"""
Modulate visual properties of Geometry objects.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from os.path import exists

from qtpy.QtCore import Signal, Qt

from mosaic.geometry import BASE_COLOR
from mosaic.utils import Throttle
from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QDialog,
    QPushButton,
    QFileDialog,
    QRadioButton,
    QWidget,
    QLabel,
    QGroupBox,
    QApplication,
)
import qtawesome as qta

from ..stylesheets import (
    QPushButton_style,
    QSpinBox_style,
    QLineEdit_style,
    QGroupBox_style,
    Colors,
)
from ..widgets import (
    DialogFooter,
    create_setting_widget,
    get_widget_value,
    ColorPickerRow,
    SliderRow,
)


class GeometryPropertiesDialog(QDialog):
    parametersChanged = Signal(dict)

    def __init__(self, initial_properties=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Properties")
        self.setFixedWidth(400)
        self.parameters = {}

        self.base_color = initial_properties.get("base_color", BASE_COLOR)
        self.highlight_color = initial_properties.get(
            "highlight_color", (0.8, 0.2, 0.2)
        )
        self.initial_properties = initial_properties or {}
        self.volume_path = self.initial_properties.get("volume_path", None)
        try:
            if not exists(self.volume_path):
                self.volume_path = None
        except Exception:
            pass

        self.setup_ui()
        self.connect_signals()

    def showEvent(self, event):
        """Position the dialog on the left side of the parent window."""
        super().showEvent(event)

        # Find the main window
        main_window = self.parent().window() if self.parent() else None
        if main_window is None:
            main_window = QApplication.activeWindow()
        if main_window is None or main_window is self:
            return

        parent_geo = main_window.geometry()

        # Position on the left side, vertically centered
        x = parent_geo.left() + 20
        y = parent_geo.top() + (parent_geo.height() - self.height()) // 2

        self.move(x, y)

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(16, 12, 16, 8)
        main_layout.setSpacing(12)

        # === APPEARANCE SECTION ===
        appearance_group = QGroupBox("Appearance")
        appearance_group.setStyleSheet(QGroupBox_style)
        appearance_layout = QVBoxLayout(appearance_group)
        appearance_layout.setSpacing(12)

        point_size_row = QWidget()
        point_size_layout = QHBoxLayout(point_size_row)
        point_size_layout.setContentsMargins(0, 0, 0, 0)
        point_size_layout.setSpacing(12)

        point_size_label = QLabel("Point Size")
        point_size_layout.addWidget(point_size_label)
        point_size_layout.addStretch()

        self.size_spin = create_setting_widget(
            {
                "type": "number",
                "min": 0,
                "max": 50,
                "default": self.initial_properties.get("size", 8),
            }
        )
        self.size_spin.setStyleSheet(QSpinBox_style)
        self.size_spin.setFixedWidth(80)
        self.size_spin.setToolTip("Size of points in the representation")
        point_size_layout.addWidget(self.size_spin)

        appearance_layout.addWidget(point_size_row)

        self.opacity_slider = SliderRow(
            "Opacity",
            min_val=0.0,
            max_val=1.0,
            default=self.initial_properties.get("opacity", 0.3),
            decimals=2,
        )
        self.opacity_slider.setToolTip(
            "Transparency of the geometry (0 = invisible, 1 = solid)"
        )
        appearance_layout.addWidget(self.opacity_slider)

        self.base_color_picker = ColorPickerRow("Base Color", self.base_color)
        self.base_color_picker.setToolTip("Default color for the geometry")
        appearance_layout.addWidget(self.base_color_picker)

        self.highlight_color_picker = ColorPickerRow(
            "Highlight Color", self.highlight_color
        )
        self.highlight_color_picker.setToolTip("Color when geometry is selected")
        appearance_layout.addWidget(self.highlight_color_picker)

        main_layout.addWidget(appearance_group)

        # === LIGHTING SECTION ===
        lighting_group = QGroupBox("Lighting")
        lighting_group.setStyleSheet(QGroupBox_style)
        lighting_layout = QVBoxLayout(lighting_group)
        lighting_layout.setSpacing(12)

        self.ambient_slider = SliderRow(
            "Ambient",
            min_val=0.0,
            max_val=1.0,
            default=self.initial_properties.get("ambient", 0.3),
            decimals=2,
        )
        self.ambient_slider.setToolTip(
            "Base illumination independent of light direction"
        )
        lighting_layout.addWidget(self.ambient_slider)

        self.diffuse_slider = SliderRow(
            "Diffuse",
            min_val=0.0,
            max_val=1.0,
            default=self.initial_properties.get("diffuse", 0.3),
            decimals=2,
        )
        self.diffuse_slider.setToolTip(
            "Scattered light reflection for a matte appearance"
        )
        lighting_layout.addWidget(self.diffuse_slider)

        self.specular_slider = SliderRow(
            "Specular",
            min_val=0.0,
            max_val=1.0,
            default=self.initial_properties.get("specular", 0.3),
            decimals=2,
        )
        self.specular_slider.setToolTip("Sharp highlights for a shiny appearance")
        lighting_layout.addWidget(self.specular_slider)

        main_layout.addWidget(lighting_group)

        # === MODEL SECTION ===
        model_group = QGroupBox("Model")
        model_group.setStyleSheet(QGroupBox_style)
        model_layout = QVBoxLayout(model_group)
        model_layout.setSpacing(12)

        browse_row = QWidget()
        browse_layout = QHBoxLayout(browse_row)
        browse_layout.setContentsMargins(0, 0, 0, 0)
        browse_layout.setSpacing(8)

        browse_label = QLabel("Map File")
        browse_label.setToolTip("Density map in MRC or CCP4 format")
        browse_layout.addWidget(browse_label)
        browse_layout.addStretch()

        self.browse_button = QPushButton("Browse...")
        self.browse_button.setIcon(qta.icon("ph.folder-open", color=Colors.ICON))
        self.browse_button.setStyleSheet(QPushButton_style)
        self.browse_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.browse_button.setToolTip("Select a density map file")
        self.browse_button.clicked.connect(self.browse_volume)
        browse_layout.addWidget(self.browse_button)

        self.attach_button = QPushButton("Reattach")
        self.attach_button.setIcon(qta.icon("ph.link", color=Colors.ICON))
        self.attach_button.setStyleSheet(QPushButton_style)
        self.attach_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.attach_button.setEnabled(self.volume_path is not None)
        self.attach_button.setToolTip("Recompute isosurface from the density map")
        browse_layout.addWidget(self.attach_button)

        model_layout.addWidget(browse_row)

        scale_row = QWidget()
        scale_layout = QHBoxLayout(scale_row)
        scale_layout.setContentsMargins(0, 0, 0, 0)
        scale_layout.setSpacing(12)

        scale_label = QLabel("Scale")
        scale_layout.addWidget(scale_label)
        scale_layout.addStretch()

        self.scale_positive = QRadioButton("+1")
        self.scale_positive.setToolTip("Use positive density values")
        self.scale_negative = QRadioButton("-1")
        self.scale_negative.setToolTip("Invert density (for negative stain maps)")
        if self.initial_properties.get("volume_scale", 0) >= 0:
            self.scale_positive.setChecked(True)
        else:
            self.scale_negative.setChecked(True)

        scale_layout.addWidget(self.scale_positive)
        scale_layout.addWidget(self.scale_negative)

        self.scale_positive.setEnabled(self.volume_path is not None)
        self.scale_negative.setEnabled(self.volume_path is not None)

        model_layout.addWidget(scale_row)

        self.isovalue_slider = SliderRow(
            "Isovalue",
            min_val=0.0,
            max_val=100.0,
            default=self.initial_properties.get("isovalue_percentile", 99.5),
            decimals=1,
            suffix="%",
            steps=1000,
            exponent=2.0,
        )
        self.isovalue_slider.setToolTip(
            "Density threshold percentile for isosurface extraction"
        )
        self.isovalue_slider.setEnabled(self.volume_path is not None)
        model_layout.addWidget(self.isovalue_slider)

        main_layout.addWidget(model_group)

        # === SAMPLING SECTION ===
        sampling_group = QGroupBox("Sampling")
        sampling_group.setStyleSheet(QGroupBox_style)
        sampling_rate = self.initial_properties.get("sampling_rate", (1.0, 1.0, 1.0))

        sampling_layout = QHBoxLayout(sampling_group)
        sampling_layout.setSpacing(8)

        base = {"type": "text", "min": 0}

        sampling_tooltip = "Voxel size in Ångström for this axis"

        min_width = 50
        sampling_layout.addWidget(QLabel("X"))
        self.sampling_x = create_setting_widget(base | {"default": sampling_rate[0]})
        self.sampling_x.setMinimumWidth(min_width)
        self.sampling_x.setStyleSheet(QLineEdit_style)
        self.sampling_x.setToolTip(sampling_tooltip)
        sampling_layout.addWidget(self.sampling_x)

        sampling_layout.addWidget(QLabel("Y"))
        self.sampling_y = create_setting_widget(base | {"default": sampling_rate[1]})
        self.sampling_y.setMinimumWidth(min_width)
        self.sampling_y.setStyleSheet(QLineEdit_style)
        self.sampling_y.setToolTip(sampling_tooltip)
        sampling_layout.addWidget(self.sampling_y)

        sampling_layout.addWidget(QLabel("Z"))
        self.sampling_z = create_setting_widget(base | {"default": sampling_rate[2]})
        self.sampling_z.setMinimumWidth(min_width)
        self.sampling_z.setStyleSheet(QLineEdit_style)
        self.sampling_z.setToolTip(sampling_tooltip)
        sampling_layout.addWidget(self.sampling_z)

        main_layout.addWidget(sampling_group)

        # Footer: Reset | [stretch] | Cancel | Done
        footer = DialogFooter(dialog=self, margin=(0, 0, 0, 0))

        self.reset_button = QPushButton("Reset")
        self.reset_button.setIcon(
            qta.icon("ph.arrow-counter-clockwise", color=Colors.ICON)
        )
        self.reset_button.setStyleSheet(QPushButton_style)
        self.reset_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.reset_button.clicked.connect(self._reset_to_defaults)
        footer.layout().insertWidget(0, self.reset_button)
        footer.layout().insertStretch(1)

        main_layout.addWidget(footer)

    def connect_signals(self):
        """Connect all widget signals to update parameters."""
        self._emit_throttle = Throttle(
            lambda *args: self.emit_parameters(), interval_ms=150
        )
        self.size_spin.valueChanged.connect(self._emit_throttle)
        self.opacity_slider.valueChanged.connect(self._emit_throttle)
        self.ambient_slider.valueChanged.connect(self._emit_throttle)
        self.diffuse_slider.valueChanged.connect(self._emit_throttle)
        self.specular_slider.valueChanged.connect(self._emit_throttle)
        self.isovalue_slider.valueChanged.connect(self._emit_throttle)
        self.scale_positive.toggled.connect(self.emit_parameters)
        self.scale_negative.toggled.connect(self.emit_parameters)
        self.sampling_x.textChanged.connect(self.emit_parameters)
        self.sampling_y.textChanged.connect(self.emit_parameters)
        self.sampling_z.textChanged.connect(self.emit_parameters)
        self.base_color_picker.colorChanged.connect(self.emit_parameters)
        self.highlight_color_picker.colorChanged.connect(self.emit_parameters)

        self.attach_button.clicked.connect(self.reattach_emit)

    def reattach_emit(self):
        parameters = self.get_parameters()
        parameters["reattach_volume"] = True
        self.parametersChanged.emit(parameters)

    def emit_parameters(self):
        parameters = self.get_parameters()
        self.parametersChanged.emit(parameters)

    def _reset_to_defaults(self):
        """Reset all values to initial properties."""
        self.size_spin.setValue(self.initial_properties.get("size", 8))
        self.opacity_slider.setValue(self.initial_properties.get("opacity", 0.3))
        self.ambient_slider.setValue(self.initial_properties.get("ambient", 0.3))
        self.diffuse_slider.setValue(self.initial_properties.get("diffuse", 0.3))
        self.specular_slider.setValue(self.initial_properties.get("specular", 0.3))

        self.base_color_picker.set_color(
            self.initial_properties.get("base_color", BASE_COLOR)
        )
        self.highlight_color_picker.set_color(
            self.initial_properties.get("highlight_color", (0.8, 0.2, 0.2))
        )

        sampling_rate = self.initial_properties.get("sampling_rate", (1.0, 1.0, 1.0))
        self.sampling_x.setText(str(sampling_rate[0]))
        self.sampling_y.setText(str(sampling_rate[1]))
        self.sampling_z.setText(str(sampling_rate[2]))

        self.emit_parameters()

    def browse_volume(self):
        from ..formats.parser import load_density

        file_name, _ = QFileDialog.getOpenFileName(
            self, "Select Volume File", "", "MRC Files (*.mrc);;All Files (*.*)"
        )
        if not file_name:
            return

        # Auto determine scale
        self.volume_path = file_name
        volume = load_density(self.volume_path)
        non_negative = (volume.data > 0).sum()
        if non_negative < volume.data.size // 2:
            self.scale_negative.setChecked(True)

        # Enable volume controls
        self.scale_positive.setEnabled(True)
        self.scale_negative.setEnabled(True)
        self.isovalue_slider.setEnabled(True)
        self.attach_button.setEnabled(True)

        self.emit_parameters()

    def get_parameters(self) -> dict:
        """Return current parameters."""
        return {
            "size": get_widget_value(self.size_spin),
            "opacity": self.opacity_slider.value(),
            "ambient": self.ambient_slider.value(),
            "diffuse": self.diffuse_slider.value(),
            "specular": self.specular_slider.value(),
            "base_color": self.base_color_picker.get_color(),
            "highlight_color": self.highlight_color_picker.get_color(),
            "scale": -1 if self.scale_negative.isChecked() else 1,
            "isovalue_percentile": self.isovalue_slider.value(),
            "volume_path": self.volume_path,
            "reattach_volume": False,
            "sampling_rate": (
                float(get_widget_value(self.sampling_x)),
                float(get_widget_value(self.sampling_y)),
                float(get_widget_value(self.sampling_z)),
            ),
        }



"""
Utility functions.

Copyright (c) 2023-2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import List, Optional, Callable

import numpy as np
from qtpy.QtCore import QTimer
from scipy import ndimage
from scipy.spatial import KDTree
from scipy.sparse import coo_matrix
from scipy.sparse.csgraph import connected_components as sparse_connected_components

__all__ = [
    "points_to_volume",
    "volume_to_points",
    "connected_components",
    "envelope_components",
    "dbscan_clustering",
    "birch_clustering",
    "eigenvalue_outlier_removal",
    "statistical_outlier_removal",
    "find_closest_points",
    "find_closest_points_cutoff",
    "com_cluster_points",
    "compute_normals",
    "compute_bounding_box",
    "cmap_to_vtkctf",
    "get_cmap",
    "normals_to_rot",
    "apply_quat",
    "NORMAL_REFERENCE",
    "skeletonize",
    "Throttle",
]

NORMAL_REFERENCE = (0, 0, 1)


class Throttle:
    """Throttle wrapper for limiting function call frequency.

    Wraps a callable to ensure it executes at most once per interval.
    The first call executes immediately; subsequent calls within the
    interval are ignored. After the interval, the next call executes.

    Parameters
    ----------
    func : Callable
        The function to throttle.
    interval_ms : int, optional
        Minimum interval between calls in milliseconds, by default 100.

    Examples
    --------
    >>> def on_slider_change(value):
    ...     print(f"Value: {value}")
    >>> throttled = Throttle(on_slider_change, interval_ms=100)
    >>> slider.valueChanged.connect(throttled)
    """

    def __init__(self, func: Callable, interval_ms: int = 100):
        self._func = func
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._timer.setInterval(interval_ms)
        self._timer.timeout.connect(self._reset)
        self._can_call = True
        self._pending_args = None
        self._pending_kwargs = None

    def __call__(self, *args, **kwargs):
        """Execute the function if throttle allows."""
        if self._can_call:
            self._can_call = False
            self._pending_args = None
            self._pending_kwargs = None
            self._timer.start()
            return self._func(*args, **kwargs)
        self._pending_args = args
        self._pending_kwargs = kwargs
        return None

    def _reset(self):
        """Reset throttle, executing any pending trailing-edge call."""
        self._can_call = True
        if self._pending_args is not None:
            args, kwargs = self._pending_args, self._pending_kwargs or {}
            self._pending_args = None
            self._pending_kwargs = None
            self(*args, **kwargs)


def points_to_volume(
    points,
    sampling_rate=1,
    shape=None,
    weight=1,
    out=None,
    use_offset: bool = False,
    out_dtype=None,
):
    """
    Convert point cloud to a volumetric representation.

    Parameters
    ----------
    points : ndarray
        Input point cloud coordinates.
    sampling_rate : float, optional
        Spacing between volume voxels, by default 1.
    shape : tuple, optional
        Output volume dimensions. If None, automatically determined from points.
    weight : float, optional
        Weight value for each individual point. Defaults to one.
    out : ndarray, optional
        Array to place result into.
    use_offset: bool
        Move points to origin and return the corresponding offset.
    out_dtype: type
        Dtype of the output array if out is not explicitly passed.

    Returns
    -------
    ndarray
        volume ndarray of point densities
    ndarray
        Array of offsets if use_offset is True.
    """
    positions = np.rint(np.divide(points, sampling_rate)).astype(int)
    if use_offset:
        offset = positions.min(axis=0)
        positions -= offset

    if shape is None:
        shape = positions.max(axis=0) + 1

    valid_mask = np.all((positions >= 0) & (positions < shape), axis=1)
    positions = positions[valid_mask]

    if out is None:
        out_dtype = np.float32 if out_dtype is None else out_dtype
        out = np.zeros(tuple(int(x) for x in shape), dtype=out_dtype)

    out[tuple(positions.T)] = weight
    if use_offset:
        return out, offset
    return out


def volume_to_points(
    volume,
    sampling_rate,
    reverse_order: bool = False,
    max_cluster: Optional[int] = None,
):
    """
    Convert volumetric segmentation to point clouds.

    Parameters
    ----------
    volume : ndarray
        Input volumetric data with cluster labels.
    sampling_rate : float
        Spacing between volume voxels.
    max_cluster : int
        Maximum number of clusters to consider before raising an error. This avoid
        accidentally loading a density volume instead of a segmentation. Default is
        no cutoff.

    Returns
    -------
    list
        List of point clouds, one for each unique cluster label.
    """
    mask = volume != 0

    # Sanity check to avoid wasting time parsing densities instead of segmentations
    if mask.sum() >= 0.7 * volume.size:
        n_points = min(50 * 50 * 50, volume.size)

        rng = np.random.default_rng()
        random_indices = rng.integers(0, volume.size, size=n_points)
        clusters = np.unique(volume.flat[random_indices])
        if max_cluster is not None and clusters.size > max_cluster:
            raise ValueError(
                f"Found {clusters.size} clusters (max: {max_cluster}). \n"
                "Make sure you are opening a segmentation."
            )

    points = np.flatnonzero(mask)
    clusters, cluster_indices = np.unique(volume.flat[points], return_inverse=True)
    if max_cluster is not None and clusters.size > max_cluster:
        raise ValueError(
            f"Found {clusters.size} clusters (max: {max_cluster}). \n"
            "Make sure you are opening a segmentation."
        )

    points = np.array(np.unravel_index(points, volume.shape), dtype=np.float32).T

    ret = []
    for index in range(len(clusters)):
        cl_points = points[cluster_indices == index]

        if reverse_order:
            indices = np.ravel_multi_index(
                cl_points[:, ::-1].T.astype(np.intp), volume.shape[::-1]
            )
            cl_points = cl_points[np.argsort(indices)]

        cl_points = np.multiply(cl_points, sampling_rate, out=cl_points)
        ret.append(cl_points)
    return ret


def _get_adjacency_matrix(points, symmetric: bool = False, eps: float = 0.0):
    # Leafsize needs to be tuned depending on the structure of the input data.
    # Points typically originates from voxel membrane segmentation on regular grids.
    # Leaf sizes between 8 - 16 work reasonably well.
    tree = KDTree(
        points,
        leafsize=16,
        compact_nodes=False,
        balanced_tree=False,
        copy_data=False,
    )
    pairs = tree.query_pairs(r=np.sqrt(3), eps=eps, output_type="ndarray")

    n_points = points.shape[0]
    adjacency = coo_matrix(
        (np.ones(len(pairs)), (pairs[:, 0], pairs[:, 1])),
        shape=(n_points, n_points),
        dtype=np.int8,
    )
    if symmetric:
        adjacency += adjacency.T
    return adjacency


def connected_components(data, **kwargs):
    """
    Find connected components in point clouds using sparse graph representations.

    Parameters
    ----------
    points : ndarray
        Input data.
    distance : tuple of float, optional
        Distance between points to be considered connected, defaults to 1.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    adjacency = _get_adjacency_matrix(data)
    return sparse_connected_components(adjacency, directed=False, return_labels=True)[1]


def envelope_components(data, **kwargs):
    """
    Find envelope of a point cloud using sparse graph representations.

    Parameters
    ----------
    data : ndarray
        Input data.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    adjacency = _get_adjacency_matrix(data, symmetric=True, eps=0.1)
    n0 = np.asarray(adjacency.sum(axis=0)).reshape(-1)

    # This is a somewhat handwavy approximation of how many neighbors
    # an envelope point should have, but appears stable in practice
    indices = np.where(n0 < (data.shape[1] ** 3 - 4))[0]
    labels = connected_components(data[indices], **kwargs)

    total_labels = np.full(data.shape[0], fill_value=-1)
    for index, label in enumerate(np.unique(labels)):
        selection = indices[labels == label]
        total_labels[selection] = index
    return total_labels


def leiden_clustering(data, resolution_parameter: float = -7.3, **kwargs):
    """
    Find Leiden partition of a point cloud using sparse graph representations.

    Parameters
    ----------
    points : ndarray
        Input data.
    resolution_parameter : float
        Log 10 of resolution parameter. Smaller values yield coarser clusters.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    import leidenalg
    import igraph as ig

    adjacency = _get_adjacency_matrix(data, eps=0.1)

    sources, targets = adjacency.nonzero()
    edges = list(zip(sources, targets))
    g = ig.Graph(n=len(data), edges=edges)
    partitions = leidenalg.find_partition(
        g, leidenalg.CPMVertexPartition, resolution_parameter=10**resolution_parameter
    )
    labels = np.full(data.shape[0], fill_value=-1)
    for index, partition in enumerate(partitions):
        labels[partition] = index
    return labels


def dbscan_clustering(data, distance=1.0, min_points=10):
    """
    Perform DBSCAN clustering on the input points.

    Parameters
    ----------
    data : ndarray
        Input data
    distance : float, optional
        Maximum distance between two samples for one to be considered as in
        the neighborhood of the other, by default 40.
    min_points : int, optional
        Minimum number of samples in a neighborhood for a point to be considered as
        a core point, by default 20.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    from sklearn.cluster import DBSCAN

    return DBSCAN(eps=distance, min_samples=min_points).fit_predict(data)


def birch_clustering(
    data, n_clusters: int = 3, threshold: float = 0.5, branching_factor: int = 50
):
    """
    Perform Birch clustering on the input points using skimage.

    Parameters
    ----------
    data : ndarray
        Input data.
    threshold: float, optional
        The radius of the subcluster obtained by merging a new sample
        and the closest subcluster should be lesser than the threshold.
        Otherwise a new subcluster is started. Setting this value to be
        very low promotes splitting and vice-versa.
    branching_factor: int, optional
        Maximum number of CF subclusters in each node. If a new samples
        enters such that the number of subclusters exceed the branching_factor
        then that node is split into two nodes with the subclusters
        redistributed in each. The parent subcluster of that node is removed
        and two new subclusters are added as parents of the 2 split nodes.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    from sklearn.cluster import Birch

    return Birch(
        n_clusters=n_clusters, threshold=threshold, branching_factor=branching_factor
    ).fit_predict(data)


def kmeans_clustering(data, k=2, **kwargs):
    """Split point cloud into k using K-means.

    Parameters
    ----------
    data : ndarray
        Input data.
    k : int
        Number of clusteres.

    Returns
    -------
    ndarray
        Cluster labels.
    """
    from sklearn.cluster import KMeans

    return KMeans(n_clusters=k, n_init="auto").fit_predict(data)


def eigenvalue_outlier_removal(points, k_neighbors=20, thresh=0.05):
    """
    Remove outliers using covariance-based edge detection.

    Parameters
    ----------
    points : ndarray
        Input point cloud.
    k_neighbors : int, optional
        Number of neighbors to consider, by default 300.
    thresh : float, optional
        Threshold for outlier detection, by default 0.05.

    Returns
    -------
    mask
        Boolean array with non-outlier points.

    References
    ----------
    .. [1]  https://github.com/denabazazian/Edge_Extraction/blob/master/Difference_Eigenvalues.py
    """
    tree = KDTree(points)
    distances, indices = tree.query(points, k=k_neighbors + 1, workers=-1)

    points_centered = points[indices[:, 1:]] - points[:, np.newaxis, :]
    cov_matrices = (
        np.einsum("ijk,ijl->ikl", points_centered, points_centered) / k_neighbors
    )

    eigenvalues = np.linalg.eigvalsh(cov_matrices)
    eigenvalues = np.sort(eigenvalues, axis=1)[:, ::-1]

    sum_eg = np.sum(eigenvalues, axis=1)
    sigma = eigenvalues[:, 0] / sum_eg

    mask = sigma >= thresh
    return mask


def statistical_outlier_removal(points, k_neighbors=20, thresh=2.0):
    """
    Remove statistical outliers from the point cloud.

    Parameters
    ----------
    points : ndarray
        Input point cloud.
    k_neighbors : int, optional
        Number of neighbors to use for mean distance estimation, by default 100.
    thresh : float, optional
        Standard deviation ratio to identify outliers, by default 0.2.

    Returns
    -------
    mask
        Boolean array with non-outlier points.
    """
    import open3d as o3d

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points.astype(np.float64))

    cl, ind = pcd.remove_statistical_outlier(nb_neighbors=k_neighbors, std_ratio=thresh)
    mask = np.zeros(points.shape[0], dtype=bool)
    mask[np.asarray(ind, dtype=int)] = 1
    return mask


def find_closest_points(positions1, positions2, k=1):
    positions1, positions2 = np.asarray(positions1), np.asarray(positions2)

    tree = KDTree(positions1)
    return tree.query(positions2, k=k)


def find_closest_points_cutoff(positions1, positions2, cutoff=1):
    positions1, positions2 = np.asarray(positions1), np.asarray(positions2)

    tree = KDTree(positions1)
    return tree.query_ball_point(positions2, cutoff)


def compute_normals(points: np.ndarray, k: int = 15, return_pcd: bool = False):
    import open3d as o3d

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.estimate_normals()
    pcd.normalize_normals()
    pcd.orient_normals_consistent_tangent_plane(k=k)
    if return_pcd:
        return pcd
    return np.asarray(pcd.normals)


def com_cluster_points(positions: np.ndarray, cutoff: float) -> np.ndarray:
    if not isinstance(positions, np.ndarray):
        positions = np.array(positions)

    if isinstance(cutoff, np.ndarray):
        cutoff = np.max(cutoff)

    tree = KDTree(positions)
    n_points = len(positions)
    unassigned = np.ones(n_points, dtype=bool)
    clusters = []

    unassigned_indices = np.where(unassigned)[0]
    while np.any(unassigned):
        seed_idx = np.random.choice(unassigned_indices)

        cluster_indices = tree.query_ball_point(positions[seed_idx], cutoff)
        cluster_indices = np.array([idx for idx in cluster_indices if unassigned[idx]])

        if len(cluster_indices) > 0:
            cluster_center = np.mean(positions[cluster_indices], axis=0)
            clusters.append(cluster_center)
            unassigned[cluster_indices] = False
            unassigned_indices = np.where(unassigned)[0]

    return np.array(clusters)


def compute_bounding_box(points: List[np.ndarray]) -> List[float]:
    if len(points) == 0:
        return (0, 0, 0)
    starts = points[0].min(axis=0)
    stops = points[0].max(axis=0)
    for point in points[1:]:
        starts_inner = point.min(axis=0)
        stops_inner = point.max(axis=0)
        starts = np.minimum(starts, starts_inner)
        stops = np.maximum(stops, stops_inner)

    return stops - starts, starts


def get_cmap(*args, **kwargs):
    from matplotlib.pyplot import get_cmap

    return get_cmap(*args, **kwargs)


def cmap_to_vtkctf(
    cmap, max_value, min_value, gamma: float = 1.0, transparent_range: bool = False
):
    """Convert a colormap to a VTK color transfer function or lookup table.

    Parameters
    ----------
    cmap : str
        Name of the colormap.
    max_value : float
        Maximum value for the color range.
    min_value : float
        Minimum value for the color range.
    gamma : float, optional
        Gamma correction factor, by default 1.0.
    transparent_range : bool, optional
        If True, returns a vtkLookupTable with transparent colors for values
        outside the specified range. By default False.

    Returns
    -------
    tuple
        (color_function, (min_value, max_value)) where color_function is either
        a vtkColorTransferFunction or vtkLookupTable depending on transparent_range.
    """
    import vtk

    if np.allclose(min_value, max_value):
        offset = 0.01 * max_value + 1e-6
        max_value += offset
        min_value -= offset

    colormap = get_cmap(cmap)

    if transparent_range:
        lut = vtk.vtkLookupTable()
        lut.SetNumberOfTableValues(256)
        lut.SetRange(min_value, max_value)

        for i in range(256):
            t = i / 255.0
            t = t ** (1 / gamma)
            r, g, b, _ = colormap(t)
            lut.SetTableValue(i, r, g, b, 1.0)

        lut.SetUseBelowRangeColor(True)
        lut.SetBelowRangeColor(0, 0, 0, 0)
        lut.SetUseAboveRangeColor(True)
        lut.SetAboveRangeColor(0, 0, 0, 0)
        lut.Build()

        return lut, (min_value, max_value)

    value_range = max_value - min_value

    # Extend color map beyond data range to avoid wrapping
    offset = value_range / 255.0
    max_value += offset

    color_transfer_function = vtk.vtkColorTransferFunction()
    for i in range(256):
        data_value = min_value + i * offset
        x = (data_value - min_value) / (max_value - min_value)
        x = max(0, min(1, x))
        x = x ** (1 / gamma)

        color_transfer_function.AddRGBPoint(data_value, *colormap(x)[0:3])

    return color_transfer_function, (min_value, max_value)


def normals_to_rot(
    normals: np.ndarray,
    target: np.ndarray = NORMAL_REFERENCE,
    mode: str = "quat",
    degrees: bool = True,
    **kwargs,
) -> np.ndarray:
    """
    Finds the shortest rotation that aligns each normal vector to its
    corresponding target vector.

    Parameters
    ----------
    normals : np.ndarray
        Input normal vectors, shape (N, 3) or (3,).
    target : np.ndarray
        Target direction(s), shape (N, 3) or (3,). If single target provided,
        it will be broadcast to all normals.
    mode : str, optional
        Output format: "quat", "matrix", or "euler". Default is "quat".
    degrees : bool, optional
        If True and mode="euler", return angles in degrees. Default is True.

    Returns
    -------
    np.ndarray
        Rotations in requested format:
        - "quat": shape (N, 4), scalar-first [w, x, y, z]
        - "matrix": shape (N, 3, 3)
        - "euler": shape (N, 3), ZYZ Euler angles
    """
    if mode not in ("quat", "matrix", "euler"):
        raise ValueError(f"Unknown mode '{mode}'. Use 'quat', 'matrix', or 'euler'.")

    normals = np.atleast_2d(normals)
    targets = np.atleast_2d(target)

    # Broadcast target if single value provided
    if targets.shape[0] == 1 and normals.shape[0] > 1:
        targets = np.repeat(targets, normals.shape[0], axis=0)
    elif targets.shape[0] != normals.shape[0]:
        raise ValueError(
            f"Incompatible shapes: normals {normals.shape}, target {targets.shape}. "
            "Provide either a single target or one per normal."
        )

    normals = normals / np.linalg.norm(normals, axis=1, keepdims=True)
    targets = targets / np.linalg.norm(targets, axis=1, keepdims=True)

    ret = _align_vectors_to_quat(targets, normals)
    if mode == "matrix":
        ret = _quat_to_matrix(ret)
    elif mode == "euler":
        ret = _quat_to_euler(ret, degrees=degrees)
    return ret


def _align_vectors_to_quat(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """
    Compute quaternions for shortest rotation aligning vec1 to vec2.

    Parameters
    ----------
    vec1, vec2 : np.ndarray
        Normalized vectors, shape (N, 3).

    Returns
    -------
    np.ndarray
        Quaternions [w, x, y, z], shape (N, 4).
    """
    axis = np.cross(vec1, vec2)
    cos_angle = np.sum(vec1 * vec2, axis=1)

    aligned = cos_angle > (1 - 1e-8)
    opposite = cos_angle < (-1 + 1e-8)
    normal = ~(aligned | opposite)

    quaternions = np.empty((vec1.shape[0], 4))
    quaternions[aligned] = [1, 0, 0, 0]

    # Half angle
    if np.any(normal):
        w = np.sqrt((1 + cos_angle[normal]) / 2)
        xyz = axis[normal] / (2 * w[:, np.newaxis])
        quaternions[normal, 0] = w
        quaternions[normal, 1:] = xyz

    # Opposite
    if np.any(opposite):
        for i in np.where(opposite)[0]:
            v = vec1[i]
            perp = np.cross(v, [1, 0, 0]) if abs(v[0]) < 0.9 else np.cross(v, [0, 1, 0])
            perp = perp / np.linalg.norm(perp)
            quaternions[i] = [0, perp[0], perp[1], perp[2]]

    return quaternions


def _quat_to_matrix(q: np.ndarray) -> np.ndarray:
    """
    Convert quaternions to rotation matrices.

    Parameters
    ----------
    q : np.ndarray
        Quaternions [w, x, y, z], shape (N, 4).

    Returns
    -------
    np.ndarray
        Rotation matrices, shape (N, 3, 3).
    """
    w, x, y, z = q[:, 0], q[:, 1], q[:, 2], q[:, 3]

    matrices = np.empty((q.shape[0], 3, 3))

    matrices[:, 0, 0] = 1 - 2 * (y**2 + z**2)
    matrices[:, 0, 1] = 2 * (x * y - w * z)
    matrices[:, 0, 2] = 2 * (x * z + w * y)

    matrices[:, 1, 0] = 2 * (x * y + w * z)
    matrices[:, 1, 1] = 1 - 2 * (x**2 + z**2)
    matrices[:, 1, 2] = 2 * (y * z - w * x)

    matrices[:, 2, 0] = 2 * (x * z - w * y)
    matrices[:, 2, 1] = 2 * (y * z + w * x)
    matrices[:, 2, 2] = 1 - 2 * (x**2 + y**2)

    return matrices


def _quat_to_euler(q: np.ndarray, degrees: bool = True) -> np.ndarray:
    """
    Convert quaternions to ZYZ Euler angles.

    Parameters
    ----------
    q : np.ndarray
        Quaternions [w, x, y, z], shape (N, 4).
    degrees : bool
        Return in degrees if True, radians if False.

    Returns
    -------
    np.ndarray
        ZYZ Euler angles [alpha, beta, gamma], shape (N, 3).
    """
    R = _quat_to_matrix(q)

    r13 = R[:, 0, 2]
    r23 = R[:, 1, 2]
    r31 = R[:, 2, 0]
    r32 = R[:, 2, 1]
    r33 = R[:, 2, 2]

    beta = np.arccos(np.clip(r33, -1, 1))
    alpha = np.arctan2(r23, r13)
    gamma = np.arctan2(r32, -r31)

    gimbal_lock = np.abs(r33) > (1 - 1e-6)

    # When r33 ≈ 1 (beta ≈ 0)
    gimbal_pos = gimbal_lock & (r33 > 0)
    alpha = np.where(gimbal_pos, np.arctan2(-R[:, 0, 1], R[:, 1, 1]), alpha)
    beta = np.where(gimbal_pos, 0, beta)
    gamma = np.where(gimbal_pos, 0, gamma)

    # When r33 ≈ -1 (beta ≈ π)
    gimbal_neg = gimbal_lock & (r33 <= 0)
    alpha = np.where(gimbal_neg, np.arctan2(R[:, 0, 1], R[:, 1, 1]), alpha)
    beta = np.where(gimbal_neg, np.pi, beta)
    gamma = np.where(gimbal_neg, 0, gamma)

    euler = np.column_stack([alpha, beta, gamma])
    return np.degrees(euler) if degrees else euler


def quat_to_euler(
    q: np.ndarray, degrees: bool = True, inv_quat: bool = False
) -> np.ndarray:
    """
    Convert quaternions to ZYZ Euler angles.

    Equivalent to scipy's
        Rotation.from_quat(q, scalar_first=True).inv().as_euler("ZYZ")
    when inv_quat=True.

    Parameters
    ----------
    q : np.ndarray
        Quaternions [w, x, y, z], shape (N, 4).
    degrees : bool
        Return in degrees if True, radians if False. Default is True.
    inv_quat : bool
        Invert quaternions before conversion. Default is False.

    Returns
    -------
    np.ndarray
        ZYZ Euler angles [alpha, beta, gamma], shape (N, 3).
    """
    if inv_quat:
        q = q.copy()
        q[:, 1:] = -q[:, 1:]

    return _quat_to_euler(q, degrees=degrees)


def apply_quat(quaternions, target=NORMAL_REFERENCE):
    return _quat_to_matrix(quaternions) @ target


def skeletonize(
    mask: np.ndarray,
    sigma: float = 1.0,
    mode: str = "core",
    batch_size: int = 100000,
) -> np.ndarray:
    """
    Skeletonize a membrane segmentation using distance transform and Hessian analysis.

    Parameters
    ----------
    mask : np.ndarray
        Binary mask where non-zero values represent structures of interest.
    sigma : float, optional
        Gaussian smoothing sigma applied to Hessian components. Default is 1.0.
    mode : {'core', 'boundary'}, optional
        Type of skeleton to extract:
        - 'core': Extract centerline/medial axis of the structure (default)
        - 'boundary': Extract skeleton along the boundary
    batch_size : int
        Number of coordinates to process per chunk. Larger values require
        more memory.

    Returns
    -------
    np.ndarray
        Binary skeleton

    References
    ----------
    .. [1]  Martinez-Sanchez, A. et al (2014) JSB,
            https://doi.org/10.1016/j.jsb.2014.02.015
    .. [2]  Lamm, L. et al. (2024) bioRxiv, doi.org/10.1101/2024.01.05.574336.

    Notes
    -----
    The original implementation is from [1]_. [2]_ adapted the code based on which
    we created this implementation, which produces very similar results but optimized
    for CPU runtime, numerical stability, and small memory footprint.
    """
    dist = -ndimage.distance_transform_edt(mask)
    if mode in ("outer", "inner"):
        mode = "boundary"

    if mode == "boundary":
        dist *= -1
    elif mode != "core":
        raise ValueError(f"mode must be 'core' or 'boundary', got '{mode}'")

    batch_size = max(int(batch_size), 1)
    grads = [ndimage.sobel(dist, axis=i) for i in range(3)]

    # Upper tri of hessian
    hessian = np.zeros((6, *dist.shape))
    hessian[0] = ndimage.gaussian_filter(ndimage.sobel(grads[0], axis=0), sigma)
    hessian[1] = ndimage.gaussian_filter(ndimage.sobel(grads[1], axis=1), sigma)
    hessian[2] = ndimage.gaussian_filter(ndimage.sobel(grads[2], axis=2), sigma)
    hessian[3] = ndimage.gaussian_filter(ndimage.sobel(grads[0], axis=1), sigma)
    hessian[4] = ndimage.gaussian_filter(ndimage.sobel(grads[0], axis=2), sigma)
    hessian[5] = ndimage.gaussian_filter(ndimage.sobel(grads[1], axis=2), sigma)

    ndim = dist.ndim
    max_eigval = np.zeros(dist.shape)
    max_eigvec = np.zeros((*dist.shape, ndim))

    coords = np.argwhere(mask)
    n_points = min(batch_size, coords.shape[0])

    H = np.zeros((n_points, ndim, ndim))
    ix = np.arange(H.shape[0], dtype=np.int32)
    for start in range(0, coords.shape[0], batch_size):
        stop = min(start + batch_size, coords.shape[0])

        x, y, z = coords[start:stop].T

        # Crop H for the last iteration otherwise reuse
        if H.shape[0] != x.shape[0]:
            H = H[: x.shape[0]]
            ix = ix[: x.shape[0]]

        H[:, 0, 0] = hessian[0, x, y, z]
        H[:, 0, 1] = hessian[3, x, y, z]
        H[:, 0, 2] = hessian[4, x, y, z]

        H[:, 1, 0] = hessian[3, x, y, z]
        H[:, 1, 1] = hessian[1, x, y, z]
        H[:, 1, 2] = hessian[5, x, y, z]

        H[:, 2, 0] = hessian[4, x, y, z]
        H[:, 2, 1] = hessian[5, x, y, z]
        H[:, 2, 2] = hessian[2, x, y, z]

        eigenvals, eigenvecs = np.linalg.eig(H)
        max_idx = np.argmax(np.abs(eigenvals), axis=-1)
        max_eigval[x, y, z] = eigenvals[ix, max_idx]
        max_eigvec[x, y, z] = eigenvecs[ix, :, max_idx]

    max_eigval = ndimage.gaussian_filter(max_eigval, sigma=1.0)
    return nonmax_suppression_trilinear(
        max_eigval, max_eigvec[..., 0], max_eigvec[..., 1], max_eigvec[..., 2], mask
    )


def nonmax_suppression_trilinear(
    values: np.ndarray,
    vx: np.ndarray,
    vy: np.ndarray,
    vz: np.ndarray,
    mask: np.ndarray,
    interp_factor: float = 0.71,
) -> np.ndarray:
    """
    Perform non-maximum suppression with trilinear interpolation along principal directions.

    Identifies local maxima by comparing each voxel's value to interpolated values along
    the forward and backward directions of its principal eigenvector. Voxels that are
    local maxima are retained in the skeleton.

    Parameters
    ----------
    values : np.ndarray
        3D array of eigenvalues at each voxel.
    vx : np.ndarray
        X-component of the principal eigenvector at each voxel.
    vy : np.ndarray
        Y-component of the principal eigenvector at each voxel.
    vz : np.ndarray
        Z-component of the principal eigenvector at each voxel.
    mask : np.ndarray
        Binary mask indicating regions of interest (1 for foreground, 0 for background).
    interp_factor : float, optional
        Interpolation distance factor along eigenvector direction. Default is 0.71.

    Returns
    -------
    np.ndarray
        Binary 3D array where True indicates skeleton voxels (local maxima).

    Notes
    -----
    A 1-voxel margin is excluded from processing to avoid edge effects. Trilinear
    interpolation is used for sub-voxel precision when comparing along eigenvector
    directions.

    References
    ----------
    .. [1]  Martinez-Sanchez, A. et al (2014) JSB,
            https://doi.org/10.1016/j.jsb.2014.02.015
    .. [2]  Lamm, L. et al. (2024) bioRxiv, doi.org/10.1101/2024.01.05.574336.
    """
    subset = tuple(slice(1, s - 1) for s in values.shape)
    inner_mask = np.zeros_like(mask)
    inner_mask[subset] = mask[subset]

    coords = np.argwhere(inner_mask)

    x, y, z = coords.T
    val = values[x, y, z]
    dx = np.abs(vx[x, y, z] * interp_factor)
    dy = np.abs(vy[x, y, z] * interp_factor)
    dz = np.abs(vz[x, y, z] * interp_factor)

    fx = x + np.sign(vx[x, y, z]).astype(int)
    fy = y + np.sign(vy[x, y, z]).astype(int)
    fz = z + np.sign(vz[x, y, z]).astype(int)

    bx = x - np.sign(vx[x, y, z]).astype(int)
    by = y - np.sign(vy[x, y, z]).astype(int)
    bz = z - np.sign(vz[x, y, z]).astype(int)

    val_forward = (
        values[x, y, z] * (1 - dx) * (1 - dy) * (1 - dz)
        + values[fx, y, z] * dx * (1 - dy) * (1 - dz)
        + values[x, fy, z] * (1 - dx) * dy * (1 - dz)
        + values[x, y, fz] * (1 - dx) * (1 - dy) * dz
        + values[fx, fy, z] * dx * dy * (1 - dz)
        + values[fx, y, fz] * dx * (1 - dy) * dz
        + values[x, fy, fz] * (1 - dx) * dy * dz
        + values[fx, fy, fz] * dx * dy * dz
    )
    val_backward = (
        values[x, y, z] * (1 - dx) * (1 - dy) * (1 - dz)
        + values[bx, y, z] * dx * (1 - dy) * (1 - dz)
        + values[x, by, z] * (1 - dx) * dy * (1 - dz)
        + values[x, y, bz] * (1 - dx) * (1 - dy) * dz
        + values[bx, by, z] * dx * dy * (1 - dz)
        + values[bx, y, bz] * dx * (1 - dy) * dz
        + values[x, by, bz] * (1 - dx) * dy * dz
        + values[bx, by, bz] * dx * dy * dz
    )
    is_max = (val > val_forward) & (val > val_backward)

    result = np.zeros_like(values, dtype=bool)
    result[x[is_max], y[is_max], z[is_max]] = True
    return result



from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Dict, Any

from vtk import vtkTransform
from qtpy.QtWidgets import QDialog

from ..stylesheets import Colors, QPushButton_style, QScrollArea_style


class BaseAnimation(ABC):
    """Base class for all animations"""

    def __init__(
        self,
        vtk_widget,
        cdata,
        volume_viewer,
        global_start_frame=0,
        enabled=True,
        name: str = "",
    ):
        self.cdata = cdata
        self.vtk_widget = vtk_widget
        self.volume_viewer = volume_viewer
        self.global_start_frame = global_start_frame

        self.name = name
        self.enabled = enabled

        self.start_frame = 0
        self.stop_frame = 100
        self.stride = 1

        self.parameters = {}
        self._init_parameters()

    @abstractmethod
    def _init_parameters(self) -> None:
        """Initialize animation-specific parameters"""
        pass

    @abstractmethod
    def get_settings(self) -> List[Dict[str, Any]]:
        """Return a list of setting definitions for the UI"""
        pass

    @abstractmethod
    def _update(self, frame: int) -> None:
        """Implementation of frame update logic"""
        pass

    @property
    def duration(self) -> int:
        """Calculate animation duration in frames"""
        return int(self.stop_frame - self.start_frame)

    def update_parameters(self, **kwargs) -> None:
        """Update parameter settings and handle associated depencies"""
        self.parameters.update(**kwargs)

    def reset(self) -> None:
        """Reset any externally-modified state before a new frame is applied.

        Called by the composer before each frame so that animations which
        modify shared state (e.g. actor opacity) can restore defaults first.
        Override in subclasses that need this; the base implementation is a no-op.
        """
        pass

    def update(self, global_frame: int) -> None:
        """Update animation state for the given global frame"""
        if not self.enabled:
            return None

        local_frame = global_frame - self.global_start_frame + self.start_frame
        if local_frame > self.stop_frame:
            return None

        if (local_frame >= self.start_frame) and (local_frame % self.stride) == 0:
            self._update(local_frame)

    def _get_rendering_context(self, return_renderer: bool = False):
        """Return the current camera instance"""
        renderer = self.vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()
        camera = renderer.GetActiveCamera()
        if return_renderer:
            return camera, renderer
        return camera

    def _get_progress(self, frame: int):
        """Compute eased progress in [0, 1] for the given local frame.

        Returns None if the animation has zero duration.
        """
        duration = self.stop_frame - self.start_frame
        if duration <= 0:
            return None
        t = (frame - self.start_frame) / duration
        return self._ease(max(0.0, min(1.0, t)))

    def _ease(self, t: float) -> float:
        """Apply easing function to progress value t in [0, 1]."""
        easing = self.parameters.get("easing", "linear")

        if easing == "ease-in":
            return t * t
        elif easing == "ease-out":
            return 1.0 - (1.0 - t) * (1.0 - t)
        elif easing == "ease-in-out":
            if t < 0.5:
                return 2.0 * t * t
            return 1.0 - (-2.0 * t + 2.0) ** 2 / 2.0
        elif easing == "instant":
            return 1.0 if t > 0 else 0.0
        return t  # linear


class TrajectoryAnimation(BaseAnimation):
    """Animation for molecular trajectories"""

    def _available_trajectories(self):
        from mosaic.geometry import GeometryTrajectory

        models = self.cdata.format_datalist("models")

        trajectories = []
        for name, obj in models:
            if isinstance(obj, GeometryTrajectory):
                trajectories.append(name)
        return trajectories

    def _get_trajectory(self, name: str):
        models = self.cdata.format_datalist("models")
        return next((x for t, x in models if t == name), None)

    def _init_parameters(self) -> None:
        trajectories = self._available_trajectories()
        if (default := self.parameters.get("trajectory")) is None:
            try:
                default = trajectories[0]
            except IndexError:
                default = None
            self.update_parameters(trajectory=default)

    def update_parameters(self, **kwargs):
        new_trajectory = kwargs.get("trajectory")
        if new_trajectory and new_trajectory != self.parameters.get("trajectory"):
            self._trajectory = self._get_trajectory(new_trajectory)
            self.start_frame = 0
            self.stop_frame = self._trajectory.frames

        return super().update_parameters(**kwargs)

    def get_settings(self) -> List[Dict[str, Any]]:
        return [
            {
                "label": "trajectory",
                "type": "select",
                "options": self._available_trajectories(),
                "default": self.parameters.get("trajectory"),
                "description": "Select trajectories to animate.",
            },
        ]

    def _update(self, frame: int) -> None:
        if not hasattr(self, "_trajectory"):
            print("No trajectory associated with object")
            return None

        self._trajectory.display_frame(frame)
        uuids = self.cdata.models._get_selected_uuids()
        if uuids:
            self.cdata.models.set_selection_by_uuid(uuids)


class VolumeAnimation(BaseAnimation):
    """Volume slicing animation"""

    def _init_parameters(self) -> None:
        self.parameters.clear()
        self.parameters["direction"] = "forward"
        self.parameters["projection"] = "Off"
        self.parameters["axis"] = "Z"
        self.parameters["hide"] = False
        self._original_visibility = None
        try:
            self.update_parameters(
                axis=self.volume_viewer.primary.orientation_selector.currentText().lower()
            )
        except Exception:
            pass

    def reset(self) -> None:
        """Restore volume visibility to the state before this animation touched it.

        Uses the VTK actor directly to avoid triggering intermediate renders
        that would cause flickering during playback.
        """
        if self._original_visibility is not None:
            self.volume_viewer.primary.slice.SetVisibility(self._original_visibility)

    def get_settings(self) -> List[Dict[str, Any]]:
        projection = [
            self.volume_viewer.primary.project_selector.itemText(i)
            for i in range(self.volume_viewer.primary.project_selector.count())
        ]
        return [
            {
                "label": "axis",
                "type": "select",
                "options": ["x", "y", "z"],
                "default": "z",
                "description": "Axis to slice over.",
            },
            {
                "label": "direction",
                "type": "select",
                "options": ["forward", "backward"],
                "description": "Direction to slice through.",
            },
            {
                "label": "projection",
                "type": "select",
                "options": projection,
                "default": self.volume_viewer.primary.orientation_selector.currentText(),
                "description": "Direction to slice through.",
            },
            {
                "label": "hide",
                "type": "select",
                "options": ["False", "True"],
                "default": str(self.parameters.get("hide", False)),
                "description": "Hide the volume during the animation.",
            },
        ]

    def update_parameters(self, **kwargs):
        if "hide" in kwargs:
            value = kwargs["hide"]
            if isinstance(value, str):
                value = value == "True"
            kwargs["hide"] = value

        new_axis = kwargs.get("axis")
        if new_axis and new_axis != self.parameters.get("axis"):
            _mapping = {"x": 0, "y": 1, "z": 2}
            shape = self.volume_viewer.primary.get_dimensions()
            self.start_frame = 0
            self.stop_frame = shape[_mapping.get(new_axis, 0)]
            kwargs["axis"] = new_axis.upper()

        # Skip the duplicate axis handling in the parent by calling grandparent
        self.parameters.update(**kwargs)

    def update(self, global_frame: int) -> None:
        """Apply hide visibility with forwards-fill semantics.

        Before the range the volume keeps whatever visibility was set by
        ``reset()`` or an earlier animation. During and after the range
        the hide state persists so that later animations can build on it.
        """
        if not self.enabled:
            return
        local_frame = global_frame - self.global_start_frame + self.start_frame
        if local_frame < self.start_frame:
            return
        local_frame = min(self.stop_frame, local_frame)
        if (local_frame >= self.start_frame) and (local_frame % self.stride) == 0:
            self._update(local_frame)

    def _update(self, frame: int) -> None:
        if self.parameters["direction"] == "backward":
            frame = self.stop_frame - frame

        viewer = self.volume_viewer.primary

        # Track original visibility so reset() can restore it
        if self._original_visibility is None:
            self._original_visibility = viewer.is_visible

        if self.parameters.get("hide", False):
            viewer.slice.SetVisibility(False)

        # We change the widgets rather than calling the underlying functions
        # to ensure the GUI is updated accordingly for interactive views
        current_orientation = viewer.get_orientation()
        if current_orientation != self.parameters["axis"]:
            viewer.orientation_selector.setCurrentText(self.parameters["axis"])

        current_state = self.volume_viewer.primary.get_projection()
        if current_state != self.parameters["projection"]:
            viewer.project_selector.setCurrentText(self.parameters["projection"])

        viewer.set_slice(frame)


class CameraAnimation(BaseAnimation):
    """Camera orbit animation with absolute positioning for proper scrubbing."""

    def _init_parameters(self) -> None:
        self.parameters.clear()
        self.parameters.update(
            {
                "axis": "y",
                "degrees": 180,
                "direction": "forward",
            }
        )
        self._last_progress = 0.0
        self.stop_frame = 180

    def get_settings(self) -> List[Dict[str, Any]]:
        return [
            {
                "label": "axis",
                "type": "select",
                "options": ["x", "y", "z"],
                "default": self.parameters.get("axis", "y"),
                "description": "Axis to rotate over.",
            },
            {
                "label": "degrees",
                "type": "float",
                "min": 0,
                "max": 360,
                "default": self.parameters.get("degrees", 180),
                "description": "Total angle to rotate over axis.",
            },
            {
                "label": "direction",
                "type": "select",
                "options": ["forward", "reverse"],
                "default": self.parameters.get("direction", "forward"),
                "description": "Direction to rotate in.",
            },
        ]

    def _update(self, frame: int) -> None:
        camera, renderer = self._get_rendering_context(return_renderer=True)

        progress = self._get_progress(frame)
        if progress is None:
            return

        total_degrees = self.parameters["degrees"]
        if self.parameters.get("direction") == "reverse":
            total_degrees = -total_degrees

        # Incremental angle since last update
        delta_angle = total_degrees * (progress - self._last_progress)
        self._last_progress = progress

        if abs(delta_angle) < 1e-10:
            return

        # Compute rotation axis from the current camera orientation
        pos = list(camera.GetPosition())
        focal = list(camera.GetFocalPoint())
        view_up = list(camera.GetViewUp())

        view_dir = [focal[i] - pos[i] for i in range(3)]
        vd_len = sum(d * d for d in view_dir) ** 0.5
        if vd_len > 0:
            view_dir = [d / vd_len for d in view_dir]

        right = [
            view_dir[1] * view_up[2] - view_dir[2] * view_up[1],
            view_dir[2] * view_up[0] - view_dir[0] * view_up[2],
            view_dir[0] * view_up[1] - view_dir[1] * view_up[0],
        ]
        r_len = sum(d * d for d in right) ** 0.5
        if r_len > 0:
            right = [d / r_len for d in right]

        axis = self.parameters["axis"]
        if axis == "x":
            rot_axis = right
        elif axis == "y":
            rot_axis = view_up
        else:  # z
            rot_axis = view_dir

        # Apply incremental rotation around focal point
        transform = vtkTransform()
        transform.Identity()
        transform.Translate(*focal)
        transform.RotateWXYZ(delta_angle, *rot_axis)
        transform.Translate(-focal[0], -focal[1], -focal[2])

        new_pos = transform.TransformPoint(pos)
        new_view_up = transform.TransformVector(view_up)

        camera.SetPosition(*new_pos)
        camera.SetViewUp(*new_view_up)
        renderer.ResetCameraClippingRange()


class ActorSelectionDialog(QDialog):
    """Dialog for selecting actors using ContainerTreeWidget."""

    def __init__(self, cdata, current_selection=None, parent=None):
        from qtpy.QtWidgets import (
            QVBoxLayout,
            QHBoxLayout,
            QTreeWidget,
            QLabel,
            QFrame,
            QPushButton,
            QGroupBox,
        )

        from mosaic.widgets import DialogFooter
        from mosaic.widgets.container_list import (
            ContainerTreeWidget,
            StyledTreeWidgetItem,
        )

        super().__init__(parent)
        self.setWindowTitle("Select Objects")
        self.resize(400, 500)
        self.setModal(True)
        self.setStyleSheet(QPushButton_style + QScrollArea_style)

        self._cdata = cdata
        self._trees = []
        self._tree_labels = []

        layout = QVBoxLayout(self)
        current_selection = set(current_selection or [])

        # Quick select buttons
        quick_group = QGroupBox("Quick Select")
        quick_layout = QHBoxLayout(quick_group)
        quick_layout.setContentsMargins(8, 8, 8, 8)
        quick_layout.setSpacing(6)

        for label, callback in [
            ("All", self._select_all),
            ("Clusters", lambda: self._select_by_type("data")),
            ("Models", lambda: self._select_by_type("models")),
        ]:
            btn = QPushButton(label)
            btn.clicked.connect(callback)
            quick_layout.addWidget(btn)

        layout.addWidget(quick_group)

        for label, data_type, interactor in [
            ("Clusters", "data", cdata.data),
            ("Models", "models", cdata.models),
        ]:
            objects = {obj.uuid: obj for _, obj in cdata.format_datalist(data_type)}
            if not objects:
                continue

            header = QLabel(label)
            header.setStyleSheet("font-weight: 500; font-size: 12px;")
            layout.addWidget(header)

            tree = ContainerTreeWidget(border=False)
            tree.tree_widget.setSelectionMode(
                QTreeWidget.SelectionMode.ExtendedSelection
            )
            self._trees.append(tree)
            self._tree_labels.append(data_type)

            state = interactor.data_list.to_state()
            uuid_to_item = {}
            for uuid, obj in objects.items():
                item = StyledTreeWidgetItem(
                    obj._meta.get("name"),
                    obj.visible,
                    {"object_id": id(obj), "data_type": data_type, **obj._meta},
                )
                uuid_to_item[uuid] = item

            tree.apply_state(state, uuid_to_item)

            for uuid, obj in objects.items():
                if uuid in uuid_to_item and id(obj) in current_selection:
                    uuid_to_item[uuid].setSelected(True)
            layout.addWidget(tree)

            separator = QFrame()
            separator.setFrameShape(QFrame.Shape.HLine)
            separator.setStyleSheet("color: #6b7280;")
            layout.addWidget(separator)

        layout.addWidget(DialogFooter(dialog=self, margin=(0, 10, 0, 0)))

    def _select_all(self):
        """Select all items in all trees."""
        for tree in self._trees:
            tree.tree_widget.selectAll()

    def _select_by_type(self, data_type: str):
        """Select only items of a specific type (data or models)."""
        for tree, label in zip(self._trees, self._tree_labels):
            if label == data_type:
                tree.tree_widget.selectAll()
            else:
                tree.tree_widget.clearSelection()

    def get_selected_objects(self):
        selected = []
        for tree in self._trees:
            selected.extend(
                item.metadata["object_id"] for item in tree.selected_items()
            )
        return selected


class VisibilityAnimation(BaseAnimation):
    """Visibility fade animation"""

    def _init_parameters(self) -> None:
        self.parameters.clear()
        self.parameters.update(
            {"start_opacity": 1.0, "target_opacity": 0.0, "easing": "instant"}
        )
        self._original_opacities = {}

    def get_settings(self) -> List[Dict[str, Any]]:
        return [
            {
                "label": "start_opacity",
                "type": "float",
                "min": 0.0,
                "max": 1.0,
                "default": self.parameters.get("start_opacity", 1.0),
                "description": "Start opacity (0.0 for invisible, 1.0 for fully visible)",
            },
            {
                "label": "target_opacity",
                "type": "float",
                "min": 0.0,
                "max": 1.0,
                "default": self.parameters.get("target_opacity", 1.0),
                "description": "Target opacity (0.0 for invisible, 1.0 for fully visible)",
            },
            {
                "label": "easing",
                "type": "select",
                "options": ["linear", "ease-in", "ease-out", "ease-in-out", "instant"],
                "default": self.parameters.get("easing", "instant"),
                "description": "Animation style (instant for immediate change)",
            },
            {
                "label": "Objects",
                "type": "button",
                "text": "Select",
                "callback": self._open_object_selection_dialog,
                "description": "Choose which objects should be affected by the animation",
            },
        ]

    def _open_object_selection_dialog(self, _checked=None):
        """Open dialog to select which objects should be affected"""
        try:
            current_selection = self.parameters.get("selected_objects", [])
            dialog = ActorSelectionDialog(
                cdata=self.cdata, current_selection=current_selection
            )
            if dialog.exec():
                selected_objects = dialog.get_selected_objects()
                self.update_parameters(selected_objects=selected_objects)

        except Exception as e:
            print(f"Error opening object selection dialog: {e}")

        return False

    def _get_actors(self):
        actors = []
        object_ids = self.parameters.get("selected_objects", [])
        try:
            all_objects = {}
            for name, obj in self.cdata.format_datalist("data"):
                all_objects[id(obj)] = obj
            for name, obj in self.cdata.format_datalist("models"):
                all_objects[id(obj)] = obj

            actors = [all_objects[x].actor for x in object_ids if x in all_objects]

        except Exception as e:
            print(f"Error getting actors for object IDs: {e}")

        return actors

    def reset(self) -> None:
        """Restore actors to the opacity they had before this animation touched them."""
        for actor in self._get_actors():
            original = self._original_opacities.get(id(actor))
            if original is not None:
                actor.GetProperty().SetOpacity(original)

    def update(self, global_frame: int) -> None:
        """Forwards-fill: apply during and after the range, skip before.

        Before the range the actor keeps whatever opacity was set by reset()
        or an earlier animation.  After the range the end state persists.
        """
        if not self.enabled:
            return
        local_frame = global_frame - self.global_start_frame + self.start_frame
        if local_frame < self.start_frame:
            return
        local_frame = min(self.stop_frame, local_frame)
        self._update(local_frame)

    def _update(self, frame: int) -> None:
        progress = self._get_progress(frame)
        if progress is None:
            return

        start_opacity = self.parameters["start_opacity"]
        target_opacity = self.parameters["target_opacity"]
        current_opacity = start_opacity + (target_opacity - start_opacity) * progress

        for actor in self._get_actors():
            if id(actor) not in self._original_opacities:
                self._original_opacities[id(actor)] = actor.GetProperty().GetOpacity()
            actor.GetProperty().SetOpacity(current_opacity)


class WaypointAnimation(BaseAnimation):
    """Animation that smoothly moves the camera between captured waypoints.

    Each waypoint stores the full camera state (position, focal point, view-up)
    so that the camera orientation is interpolated correctly along the path.
    """

    def _init_parameters(self) -> None:
        self.parameters.clear()
        self.parameters.update(
            {
                "waypoints": [],
                "spline_order": 3,
                "easing": "linear",
            }
        )
        self._path_actors = []
        self._curve = None
        self._positions = None
        self._focal_points = None
        self._view_ups = None

    def update_parameters(self, **kwargs):
        if "add_waypoint" in kwargs:
            self._capture_waypoint()
            kwargs.pop("add_waypoint")

        if "remove_waypoint" in kwargs:
            self._remove_last_waypoint()
            kwargs.pop("remove_waypoint")

        if "clear_waypoints" in kwargs:
            self._clear_waypoints()
            kwargs.pop("clear_waypoints")

        if "waypoints" in kwargs:
            self.parameters["waypoints"] = kwargs.pop("waypoints")
            self._init_spline()

        if "spline_order" in kwargs:
            self.parameters["spline_order"] = int(kwargs.pop("spline_order"))
            self._init_spline()
            self._update_path_preview()

        if "easing" in kwargs:
            self.parameters["easing"] = kwargs.pop("easing")

        return super().update_parameters(**kwargs)

    def _capture_waypoint(self):
        """Capture the current camera state as a new waypoint."""
        camera = self._get_rendering_context()
        waypoint = {
            "position": list(camera.GetPosition()),
            "focal_point": list(camera.GetFocalPoint()),
            "view_up": list(camera.GetViewUp()),
        }
        self.parameters["waypoints"].append(waypoint)
        self._init_spline()
        self._update_path_preview()

    def _remove_last_waypoint(self):
        """Remove the most recently added waypoint."""
        waypoints = self.parameters["waypoints"]
        if waypoints:
            waypoints.pop()
            self._init_spline()
            self._update_path_preview()

    def _clear_waypoints(self):
        """Remove all waypoints."""
        self.parameters["waypoints"].clear()
        self._init_spline()
        self.cleanup_preview()

    def _init_spline(self):
        """Initialize the spline curve from waypoints."""
        import numpy as np
        from mosaic.parametrization import SplineCurve

        waypoints = self.parameters.get("waypoints", [])
        if len(waypoints) < 2:
            self._curve = None
            self._positions = None
            self._focal_points = None
            self._view_ups = None
            return

        positions = np.array([w["position"] for w in waypoints])
        focal_points = np.array([w["focal_point"] for w in waypoints])
        view_ups = np.array([w["view_up"] for w in waypoints])

        order = int(self.parameters.get("spline_order", 3))
        order = min(order, len(waypoints) - 1)

        self._curve = SplineCurve(positions=positions, order=order)

        n_samples = max(self.stop_frame - self.start_frame + 1, 2)
        self._positions = self._curve.sample(n_samples)

        # Linear interpolation for focal points and view-up vectors
        t = np.linspace(0, 1, n_samples)
        t_wp = np.linspace(0, 1, len(waypoints))
        self._focal_points = np.column_stack(
            [np.interp(t, t_wp, focal_points[:, i]) for i in range(3)]
        )
        self._view_ups = np.column_stack(
            [np.interp(t, t_wp, view_ups[:, i]) for i in range(3)]
        )

    def _update_path_preview(self):
        """Visualize the camera path in the 3D viewport.

        Draws from the same sampled positions used by the animation so the
        preview exactly matches what plays back.
        """
        import vtk

        _, renderer = self._get_rendering_context(return_renderer=True)

        # Remove old preview actors
        self.cleanup_preview()

        waypoints = self.parameters.get("waypoints", [])

        # Show point markers even for a single waypoint
        if len(waypoints) < 2 or self._positions is None:
            for wp in waypoints:
                actor = self._create_waypoint_marker(wp["position"])
                renderer.AddActor(actor)
                self._path_actors.append(actor)
            self.vtk_widget.GetRenderWindow().Render()
            return

        # Build a polyline from the sampled animation positions
        vtk_points = vtk.vtkPoints()
        for pos in self._positions:
            vtk_points.InsertNextPoint(pos)

        lines = vtk.vtkCellArray()
        n_pts = len(self._positions)
        lines.InsertNextCell(n_pts)
        for i in range(n_pts):
            lines.InsertCellPoint(i)

        poly_data = vtk.vtkPolyData()
        poly_data.SetPoints(vtk_points)
        poly_data.SetLines(lines)

        tube_filter = vtk.vtkTubeFilter()
        tube_filter.SetInputData(poly_data)
        tube_filter.SetRadius(1.0)
        tube_filter.SetNumberOfSides(8)
        tube_filter.SetVaryRadiusToVaryRadiusOff()
        tube_filter.Update()

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(tube_filter.GetOutputPort())
        mapper.SetResolveCoincidentTopologyToPolygonOffset()
        mapper.SetScalarVisibility(False)

        path_actor = vtk.vtkActor()
        path_actor.SetMapper(mapper)
        # Pink #ec4899 → RGB
        path_actor.GetProperty().SetColor(0.925, 0.286, 0.6)
        path_actor.GetProperty().SetOpacity(0.7)
        renderer.AddActor(path_actor)
        self._path_actors.append(path_actor)

        # Add sphere markers at each waypoint
        for wp in waypoints:
            actor = self._create_waypoint_marker(wp["position"])
            renderer.AddActor(actor)
            self._path_actors.append(actor)

        self.vtk_widget.GetRenderWindow().Render()

    def _create_waypoint_marker(self, position):
        """Create a small sphere actor at the given position."""
        import vtk

        point_data = vtk.vtkPoints()
        point_data.InsertNextPoint(position)

        vertices = vtk.vtkCellArray()
        vertices.InsertNextCell(1)
        vertices.InsertCellPoint(0)

        poly_data = vtk.vtkPolyData()
        poly_data.SetPoints(point_data)
        poly_data.SetVerts(vertices)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(poly_data)
        mapper.SetResolveCoincidentTopologyToPolygonOffset()

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        # Pink #ec4899
        actor.GetProperty().SetColor(0.925, 0.286, 0.6)
        actor.GetProperty().SetPointSize(12)
        actor.GetProperty().SetRenderPointsAsSpheres(True)

        return actor

    def cleanup_preview(self):
        """Remove all preview actors from the renderer."""
        try:
            _, renderer = self._get_rendering_context(return_renderer=True)
            for actor in self._path_actors:
                renderer.RemoveActor(actor)
            self._path_actors.clear()
            self.vtk_widget.GetRenderWindow().Render()
        except Exception:
            self._path_actors.clear()

    def get_settings(self) -> List[Dict[str, Any]]:
        n = len(self.parameters.get("waypoints", []))
        count_text = f"{n} waypoint{'s' if n != 1 else ''} defined"
        return [
            {
                "label": "Capture Current View",
                "type": "button",
                "text": "Capture",
                "callback": lambda _=None: self.update_parameters(add_waypoint=True),
                "description": "Add the current camera position as a waypoint",
            },
            {
                "label": "Remove Last",
                "type": "button",
                "text": "Remove",
                "callback": lambda _=None: self.update_parameters(remove_waypoint=True),
                "description": "Remove the last waypoint",
            },
            {
                "label": "Clear All",
                "type": "button",
                "text": "Clear",
                "callback": lambda _=None: self.update_parameters(clear_waypoints=True),
                "description": "Remove all waypoints",
            },
            {
                "label": "waypoints_count",
                "type": "label",
                "text": count_text,
                "description": "Number of waypoints defined",
            },
            {
                "label": "spline_order",
                "type": "select",
                "options": ["1", "2", "3"],
                "default": str(self.parameters.get("spline_order", 3)),
                "description": "Spline interpolation order (1=linear, 2=quadratic, 3=cubic)",
            },
            {
                "label": "easing",
                "type": "select",
                "options": ["linear", "ease-in", "ease-out", "ease-in-out"],
                "default": self.parameters.get("easing", "linear"),
                "description": "Easing function for playback speed",
            },
        ]

    def _update(self, frame: int) -> None:
        import numpy as np

        if len(self.parameters.get("waypoints", [])) < 2:
            return

        # (Re)sample if needed
        n_samples = self.stop_frame - self.start_frame + 1
        if self._positions is None or len(self._positions) != n_samples:
            self._init_spline()
            if self._positions is None:
                return

        progress = self._get_progress(frame)
        if progress is None:
            return

        # Map eased progress to sample index
        idx = progress * (n_samples - 1)
        idx_low = int(idx)
        idx_high = min(idx_low + 1, n_samples - 1)
        frac = idx - idx_low

        # Linearly interpolate between adjacent samples
        new_pos = (1 - frac) * self._positions[idx_low] + frac * self._positions[
            idx_high
        ]
        new_focal = (1 - frac) * self._focal_points[
            idx_low
        ] + frac * self._focal_points[idx_high]
        new_up = (1 - frac) * self._view_ups[idx_low] + frac * self._view_ups[idx_high]

        # Normalize view-up vector
        up_len = np.linalg.norm(new_up)
        if up_len > 0:
            new_up = new_up / up_len

        camera, renderer = self._get_rendering_context(return_renderer=True)
        camera.SetPosition(*new_pos)
        camera.SetFocalPoint(*new_focal)
        camera.SetViewUp(*new_up)
        renderer.ResetCameraClippingRange()


class ZoomAnimation(BaseAnimation):
    """Camera zoom animation"""

    def _init_parameters(self) -> None:
        self.parameters.clear()
        self.parameters.update(
            {
                "zoom_factor": 2.0,
                "easing": "ease-in-out",
            }
        )
        self._initial_distance = None

    def get_settings(self) -> List[Dict[str, Any]]:
        return [
            {
                "label": "zoom_factor",
                "type": "float",
                "min": 0.1,
                "max": 10.0,
                "default": self.parameters.get("zoom_factor", 2.0),
                "description": "Target zoom factor (>1 zooms in, <1 zooms out)",
            },
            {
                "label": "easing",
                "type": "select",
                "options": ["linear", "ease-in", "ease-out", "ease-in-out"],
                "default": self.parameters.get("easing", "ease-in-out"),
                "description": "Easing function for smooth zoom",
            },
        ]

    def _update(self, frame: int) -> None:
        camera, renderer = self._get_rendering_context(return_renderer=True)

        if frame == self.start_frame or self._initial_distance is None:
            self._initial_distance = camera.GetDistance()
            self._initial_position = camera.GetPosition()
            self._initial_focal = camera.GetFocalPoint()

        progress = self._get_progress(frame)
        if progress is None:
            return

        zoom_factor = self.parameters["zoom_factor"]
        target_distance = self._initial_distance / zoom_factor

        current_distance = (
            self._initial_distance
            + (target_distance - self._initial_distance) * progress
        )

        # Move camera along the view direction
        direction = [
            self._initial_position[i] - self._initial_focal[i] for i in range(3)
        ]
        length = sum(d * d for d in direction) ** 0.5
        if length > 0:
            direction = [d / length for d in direction]

        new_position = [
            self._initial_focal[i] + direction[i] * current_distance for i in range(3)
        ]

        camera.SetPosition(*new_position)
        renderer.ResetCameraClippingRange()


class AnimationType(Enum):
    TRAJECTORY = {
        "name": "Trajectory",
        "color": Colors.CATEGORY["trajectory"],
        "class": TrajectoryAnimation,
    }
    CAMERA = {
        "name": "Orbit",
        "color": Colors.CATEGORY["camera"],
        "class": CameraAnimation,
    }
    ZOOM = {"name": "Zoom", "color": Colors.CATEGORY["zoom"], "class": ZoomAnimation}
    SLICE = {
        "name": "Volume",
        "color": Colors.CATEGORY["volume"],
        "class": VolumeAnimation,
    }
    VISIBILITY = {
        "name": "Visibility",
        "color": Colors.CATEGORY["visibility"],
        "class": VisibilityAnimation,
    }
    WAYPOINT = {
        "name": "Waypoint",
        "color": Colors.CATEGORY["waypoint"],
        "class": WaypointAnimation,
    }
